from ._base import *
from .models import Params, component_name
from .exceptions import *
from .dependencies import fix_query_dependencies, clone_dependant, insert_dependencies

async def call_sync_async(call, *args, **kwargs):
    is_coroutine = asyncio.iscoroutinefunction(call)
    if is_coroutine:
        return await call(*args, **kwargs)
    else:
        return await run_in_threadpool(call, *args, **kwargs)

@component_name(f'_Request')
class JsonRpcRequest(BaseModel):
    jsonrpc: StrictStr = Field('2.0', const=True, example='2.0')
    id: Union[StrictStr, int] = Field(None, example=0)
    method: StrictStr
    params: dict

    class Config:
        extra = 'forbid'


@component_name(f'_Response')
class JsonRpcResponse(BaseModel):
    jsonrpc: StrictStr = Field('2.0', const=True, example='2.0')
    id: Union[StrictStr, int] = Field(None, example=0)
    result: dict

    class Config:
        extra = 'forbid'

def make_request_model(name, module, body_params: List[ModelField]):
    whole_params_list = [p for p in body_params if isinstance(p.field_info, Params)]
    if len(whole_params_list):
        if len(whole_params_list) > 1:
            raise RuntimeError(
                f"Only one 'Params' allowed: "
                f"params={whole_params_list}"
            )
        body_params_list = [p for p in body_params if not isinstance(p.field_info, Params)]
        if body_params_list:
            raise RuntimeError(
                f"No other params allowed when 'Params' used: "
                f"params={whole_params_list}, other={body_params_list}"
            )

    if whole_params_list:
        assert whole_params_list[0].alias == 'params'
        params_field = whole_params_list[0]
    else:
        _JsonRpcRequestParams = ModelMetaclass.__new__(ModelMetaclass, '_JsonRpcRequestParams', (BaseModel,), {})

        for f in body_params:
            _JsonRpcRequestParams.__fields__[f.name] = f

        _JsonRpcRequestParams = component_name(f'_Params[{name}]', module)(_JsonRpcRequestParams)

        params_field = ModelField(
            name='params',
            type_=_JsonRpcRequestParams,
            class_validators={},
            default=None,
            required=True,
            model_config=BaseConfig,
            field_info=Field(...),
        )

    class _Request(BaseModel):
        jsonrpc: StrictStr = Field('2.0', const=True, example='2.0')
        id: Union[StrictStr, int] = Field(None, example=0)
        method: StrictStr = Field(name, const=True, example=name)

        class Config:
            extra = 'forbid'

    _Request.__fields__[params_field.name] = params_field
    _Request = component_name(f'_Request[{name}]', module)(_Request)

    return _Request



class JsonRpcContext:
    def __init__(
        self,
        entrypoint: 'Entrypoint',
        raw_request: Any,
        http_request: Request,
        background_tasks: BackgroundTasks,
        http_response: Response,
        json_rpc_request_class: Type[JsonRpcRequest] = JsonRpcRequest,
        method_route: Optional['MethodRoute'] = None,
    ):
        self.entrypoint: Entrypoint = entrypoint
        self.raw_request: Any = raw_request
        self.http_request: Request = http_request
        self.background_tasks: BackgroundTasks = background_tasks
        self.http_response: Response = http_response
        self.request_class: Type[JsonRpcRequest] = json_rpc_request_class
        self.method_route: Optional[MethodRoute] = method_route
        self._raw_response: Optional[dict] = None
        self.exception: Optional[Exception] = None
        self.is_unhandled_exception: bool = False
        self.exit_stack: Optional[AsyncExitStack] = None
        self.jsonrpc_context_token: Optional[contextvars.Token] = None

    def on_raw_response(
        self,
        raw_response: dict,
        exception: Optional[Exception] = None,
        is_unhandled_exception: bool = False,
    ):
        raw_response.pop('id', None)
        if isinstance(self.raw_request, dict) and 'id' in self.raw_request:
            raw_response['id'] = self.raw_request.get('id')
        elif 'error' in raw_response:
            raw_response['id'] = None
        self._raw_response = raw_response
        self.exception = exception
        self.is_unhandled_exception = is_unhandled_exception

    @property
    def raw_response(self) -> dict:
        return self._raw_response

    @raw_response.setter
    def raw_response(self, value: dict):
        self.on_raw_response(value)

    @cached_property
    def request(self) -> JsonRpcRequest:
        try:
            return self.request_class.validate(self.raw_request)
        except DictError:
            raise InvalidRequest(data={'errors': [{
                'loc': (),
                'type': 'type_error.dict',
                'msg': "value is not a valid dict",
            }]})
        except ValidationError as exc:
            raise invalid_request_from_validation_error(exc)

    async def __aenter__(self):
        assert self.exit_stack is None
        self.exit_stack = await AsyncExitStack().__aenter__()
        if sentry_sdk is not None:
            self.exit_stack.enter_context(self._fix_sentry_scope())
        await self.exit_stack.enter_async_context(self._handle_exception(reraise=False))
        self.jsonrpc_context_token = _jsonrpc_context.set(self)
        return self

    async def __aexit__(self, *exc_details):
        assert self.jsonrpc_context_token is not None
        _jsonrpc_context.reset(self.jsonrpc_context_token)
        return await self.exit_stack.__aexit__(*exc_details)

    @asynccontextmanager
    async def _handle_exception(self, reraise=True):
        try:
            yield
        except Exception as exc:
            if exc is not self.exception:
                try:
                    resp, is_unhandled_exception = await self.entrypoint.handle_exception_to_resp(
                        exc, log_unhandled_exception=False,
                    )
                except Exception as exc:
                    # HTTPException
                    self._raw_response = None
                    self.exception = exc
                    self.is_unhandled_exception = True
                    raise
                self.on_raw_response(resp, exc, is_unhandled_exception)
            if reraise:
                raise

        if self.exception is not None and self.is_unhandled_exception:
            logger.exception(str(self.exception), exc_info=self.exception)

    def _make_sentry_event_processor(self):
        def event_processor(event, _):
            if self.method_route is not None:
                event['transaction'] = sentry_transaction_from_function(self.method_route.func)
            return event

        return event_processor

    @contextmanager
    def _fix_sentry_scope(self):
        hub = sentry_sdk.Hub.current
        with sentry_sdk.Hub(hub) as hub:
            with hub.configure_scope() as scope:
                scope.clear_breadcrumbs()
                scope.add_event_processor(self._make_sentry_event_processor())
                yield

    async def enter_middlewares(self, middlewares: Sequence['JsonRpcMiddleware']):
        for mw in middlewares:
            cm = mw(self)
            if not isinstance(cm, AbstractAsyncContextManager):
                raise RuntimeError("JsonRpcMiddleware(context) must return AsyncContextManager")
            await self.exit_stack.enter_async_context(cm)
            await self.exit_stack.enter_async_context(self._handle_exception())


JsonRpcMiddleware = Callable[[JsonRpcContext], AbstractAsyncContextManager]


_jsonrpc_context = contextvars.ContextVar('_lazyrpc__jsonrpc_context')


def get_jsonrpc_context() -> JsonRpcContext:
    return _jsonrpc_context.get()


def get_jsonrpc_request_id() -> Optional[Union[str, int]]:
    return get_jsonrpc_context().raw_request.get('id')


def get_jsonrpc_method() -> Optional[str]:
    return get_jsonrpc_context().raw_request.get('method')


class MethodRoute(APIRoute):
    def __init__(
        self,
        entrypoint: 'Entrypoint',
        path: str,
        func: Union[FunctionType, CoroutineType],
        *,
        result_model: Type[Any] = None,
        name: str = None,
        errors: Sequence[Type[BaseError]] = None,
        dependencies: Sequence[Depends] = None,
        response_class: Type[Response] = JSONResponse,
        request_class: Type[JsonRpcRequest] = JsonRpcRequest,
        middlewares: Sequence[JsonRpcMiddleware] = None,
        **kwargs,
    ):
        name = name or func.__name__
        result_model = result_model or func.__annotations__.get('return')

        _, path_format, _ = compile_path(path)
        func_dependant = get_dependant(path=path_format, call=func)
        insert_dependencies(func_dependant, dependencies)
        insert_dependencies(func_dependant, entrypoint.common_dependencies)
        fix_query_dependencies(func_dependant)
        flat_dependant = get_flat_dependant(func_dependant, skip_repeats=True)

        _Request = make_request_model(name, func.__module__, flat_dependant.body_params)

        @component_name(f'_Response[{name}]', func.__module__)
        class _Response(BaseModel):
            jsonrpc: StrictStr = Field('2.0', const=True, example='2.0')
            id: Union[StrictStr, int] = Field(None, example=0)
            result: result_model

            class Config:
                extra = 'forbid'

        # Only needed to generate OpenAPI
        async def endpoint(__request__: _Request):
            del __request__

        endpoint.__name__ = func.__name__
        endpoint.__doc__ = func.__doc__

        responses = errors_responses(errors)

        super().__init__(
            path,
            endpoint,
            methods=['POST'],
            name=name,
            response_class=response_class,
            response_model=_Response,
            responses=responses,
            **kwargs,
        )

        # Add dependencies and other parameters from func_dependant for correct OpenAPI generation
        self.dependant.path_params = func_dependant.path_params
        self.dependant.header_params = func_dependant.header_params
        self.dependant.cookie_params = func_dependant.cookie_params
        self.dependant.dependencies = func_dependant.dependencies
        self.dependant.security_requirements = func_dependant.security_requirements

        self.func = func
        self.func_dependant = func_dependant
        self.entrypoint = entrypoint
        self.middlewares = middlewares or []
        self.app = request_response(self.handle_http_request)
        self.request_class = request_class

    async def parse_body(self, http_request) -> Any:
        try:
            req = await http_request.json()
        except JSONDecodeError:
            raise ParseError()
        return req

    async def handle_http_request(self, http_request: Request):
        background_tasks = BackgroundTasks()

        # noinspection PyTypeChecker
        sub_response = Response(
            content=None,
            status_code=None,
            headers=None,
            media_type=None,
            background=None,
        )

        try:
            body = await self.parse_body(http_request)
        except Exception as exc:
            resp, _ = await self.entrypoint.handle_exception_to_resp(exc)
            response = self.response_class(content=resp, background=background_tasks)
        else:
            try:
                resp = await self.handle_body(http_request, background_tasks, sub_response, body)
            except NoContent:
                # no content for successful notifications
                response = Response(media_type='application/json', background=background_tasks)
            else:
                response = self.response_class(content=resp, background=background_tasks)

        response.headers.raw.extend(sub_response.headers.raw)
        if sub_response.status_code:
            response.status_code = sub_response.status_code

        return response

    async def handle_body(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        body: Any,
    ) -> dict:
        # Shared dependencies for all requests in one json-rpc batch request
        shared_dependencies_error = None
        try:
            dependency_cache = await self.entrypoint.solve_shared_dependencies(
                http_request,
                background_tasks,
                sub_response,
            )
        except BaseError as error:
            shared_dependencies_error = error
            dependency_cache = None

        resp = await self.handle_req_to_resp(
            http_request, background_tasks, sub_response, body,
            dependency_cache=dependency_cache,
            shared_dependencies_error=shared_dependencies_error,
        )

        # No response for successful notifications
        has_content = 'error' in resp or 'id' in resp
        if not has_content:
            raise NoContent

        return resp

    async def handle_req_to_resp(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        req: Any,
        dependency_cache: dict = None,
        shared_dependencies_error: BaseError = None
    ) -> dict:
        async with JsonRpcContext(
            entrypoint=self.entrypoint,
            method_route=self,
            raw_request=req,
            http_request=http_request,
            background_tasks=background_tasks,
            http_response=sub_response,
            json_rpc_request_class=self.request_class,
        ) as ctx:
            await ctx.enter_middlewares(self.entrypoint.middlewares)

            if ctx.request.method != self.name:
                raise MethodNotFound

            resp = await self.handle_req(
                http_request, background_tasks, sub_response, ctx,
                dependency_cache=dependency_cache,
                shared_dependencies_error=shared_dependencies_error,
            )
            ctx.on_raw_response(resp)

        return ctx.raw_response

    async def handle_req(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        ctx: JsonRpcContext,
        dependency_cache: dict = None,
        shared_dependencies_error: BaseError = None
    ):
        await ctx.enter_middlewares(self.middlewares)

        if shared_dependencies_error:
            raise shared_dependencies_error

        # dependency_cache - there are shared dependencies, we pass them to each method, since
        # they are common to all methods in the batch.
        # But if the methods have their own dependencies, they are resolved separately.
        dependency_cache = dependency_cache.copy()

        values, errors, background_tasks, _, _ = await solve_dependencies(
            request=http_request,
            dependant=self.func_dependant,
            body=ctx.request.params,
            background_tasks=background_tasks,
            response=sub_response,
            dependency_overrides_provider=self.dependency_overrides_provider,
            dependency_cache=dependency_cache,
        )

        if errors:
            raise invalid_params_from_validation_error(RequestValidationError(errors))

        result = await call_sync_async(self.func, **values)

        response = {
            'jsonrpc': '2.0',
            'result': result,
        }

        # noinspection PyTypeChecker
        resp = await serialize_response(
            field=self.secure_cloned_response_field,
            response_content=response,
            include=self.response_model_include,
            exclude=self.response_model_exclude,
            by_alias=self.response_model_by_alias,
            exclude_unset=self.response_model_exclude_unset,
        )

        return resp


class RequestShadow(Request):
    def __init__(self, request: Request):
        super().__init__(scope=ChainMap({}, request.scope))
        self.request = request

    async def stream(self):
        async for body in self.request.stream():
            yield body

    async def body(self):
        return await self.request.body()

    async def json(self):
        return await self.request.json()

    async def form(self):
        return await self.request.form()

    async def close(self):
        raise NotImplementedError

    async def is_disconnected(self):
        return await self.request.is_disconnected()


class EntrypointRoute(APIRoute):
    def __init__(
        self,
        entrypoint: 'Entrypoint',
        path: str,
        *,
        name: str = None,
        errors: Sequence[Type[BaseError]] = None,
        common_dependencies: Sequence[Depends] = None,
        response_class: Type[Response] = JSONResponse,
        request_class: Type[JsonRpcRequest] = JsonRpcRequest,
        **kwargs,
    ):
        name = name or 'entrypoint'

        _, path_format, _ = compile_path(path)

        _Request = request_class

        common_dependant = Dependant(path=path_format)
        if common_dependencies:
            insert_dependencies(common_dependant, common_dependencies)
            fix_query_dependencies(common_dependant)
            common_dependant = get_flat_dependant(common_dependant, skip_repeats=True)

            if common_dependant.body_params:
                _Request = make_request_model(name, entrypoint.callee_module, common_dependant.body_params)

        # This is only necessary for generating OpenAPI
        def endpoint(__request__: _Request):
            del __request__

        responses = errors_responses(errors)

        super().__init__(
            path,
            endpoint,
            methods=['POST'],
            name=name,
            response_class=response_class,
            response_model=JsonRpcResponse,
            responses=responses,
            **kwargs,
        )

        flat_dependant = get_flat_dependant(self.dependant, skip_repeats=True)

        if len(flat_dependant.body_params) > 1:
            body_params = [p for p in flat_dependant.body_params if p.type_ is not _Request]
            raise RuntimeError(
                f"Entrypoint shared dependencies can't use 'Body' parameters: "
                f"params={body_params}"
            )

        if flat_dependant.query_params:
            raise RuntimeError(
                f"Entrypoint shared dependencies can't use 'Query' parameters: "
                f"params={flat_dependant.query_params}"
            )

        self.shared_dependant = clone_dependant(self.dependant)

        # No shared 'Body' params, because each JSON-RPC request in batch has own body
        self.shared_dependant.body_params = []

        # Add dependencies and other parameters from common_dependant for correct OpenAPI generation
        self.dependant.path_params.extend(common_dependant.path_params)
        self.dependant.header_params.extend(common_dependant.header_params)
        self.dependant.cookie_params.extend(common_dependant.cookie_params)
        self.dependant.dependencies.extend(common_dependant.dependencies)
        self.dependant.security_requirements.extend(common_dependant.security_requirements)

        self.app = request_response(self.handle_http_request)
        self.entrypoint = entrypoint
        self.common_dependencies = common_dependencies
        self.request_class = request_class


    async def solve_shared_dependencies(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
    ) -> dict:
        # Must not be empty, otherwise FastAPI re-creates it
        dependency_cache = {(lambda: None, ('', )): 1}
        if self.dependencies:
            _, errors, _, _, _ = await solve_dependencies(
                request=http_request,
                dependant=self.shared_dependant,
                body=None,
                background_tasks=background_tasks,
                response=sub_response,
                dependency_overrides_provider=self.dependency_overrides_provider,
                dependency_cache=dependency_cache,
            )
            if errors:
                raise invalid_params_from_validation_error(RequestValidationError(errors))
        return dependency_cache

    async def parse_body(self, http_request) -> Any:
        try:
            body = await http_request.json()
        except JSONDecodeError:
            raise ParseError()

        if isinstance(body, list) and not body:
            raise InvalidRequest(data={'errors': [
                {'loc': (), 'type': 'value_error.empty', 'msg': "rpc call with an empty array"}
            ]})

        return body

    async def handle_http_request(self, http_request: Request):
        background_tasks = BackgroundTasks()

        # noinspection PyTypeChecker
        sub_response = Response(
            content=None,
            status_code=None,
            headers=None,
            media_type=None,
            background=None,
        )

        try:
            body = await self.parse_body(http_request)
        except Exception as exc:
            resp, _ = await self.entrypoint.handle_exception_to_resp(exc)
            response = self.response_class(content=resp, background=background_tasks)
        else:
            try:
                resp = await self.handle_body(http_request, background_tasks, sub_response, body)
            except NoContent:
                # no content for successful notifications
                response = Response(media_type='application/json', background=background_tasks)
            else:
                response = self.response_class(content=resp, background=background_tasks)

        response.headers.raw.extend(sub_response.headers.raw)
        if sub_response.status_code:
            response.status_code = sub_response.status_code

        return response

    async def handle_body(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        body: Any,
    ) -> dict:
        # Shared dependencies for all requests in one json-rpc batch request
        shared_dependencies_error = None
        try:
            dependency_cache = await self.solve_shared_dependencies(
                http_request,
                background_tasks,
                sub_response,
            )
        except BaseError as error:
            shared_dependencies_error = error
            dependency_cache = None

        scheduler = await self.entrypoint.get_scheduler()

        if isinstance(body, list):
            req_list = body
        else:
            req_list = [body]

        job_list = []
        if len(req_list) > 1:
            # Run concurrently through scheduler
            for req in req_list:
                job = await scheduler.spawn(
                    self.handle_req_to_resp(
                        http_request, background_tasks, sub_response, req,
                        dependency_cache=dependency_cache,
                        shared_dependencies_error=shared_dependencies_error,
                    )
                )

                # TODO: https://github.com/aio-libs/aiojobs/issues/119
                job._explicit = True
                # noinspection PyProtectedMember
                coro = job._do_wait(timeout=None)

                job_list.append(coro)
        else:
            req = req_list[0]
            coro = self.handle_req_to_resp(
                http_request, background_tasks, sub_response, req,
                dependency_cache=dependency_cache,
                shared_dependencies_error=shared_dependencies_error,
            )
            job_list.append(coro)

        resp_list = []

        for coro in job_list:
            resp = await coro

            # No response for successful notifications
            has_content = 'error' in resp or 'id' in resp
            if not has_content:
                continue

            resp_list.append(resp)

        if not resp_list:
            raise NoContent

        if not isinstance(body, list):
            content = resp_list[0]
        else:
            content = resp_list

        return content

    async def handle_req_to_resp(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        req: Any,
        dependency_cache: dict = None,
        shared_dependencies_error: BaseError = None
    ) -> dict:
        async with JsonRpcContext(
            entrypoint=self.entrypoint,
            raw_request=req,
            http_request=http_request,
            background_tasks=background_tasks,
            http_response=sub_response,
            json_rpc_request_class=self.request_class
        ) as ctx:
            await ctx.enter_middlewares(self.entrypoint.middlewares)

            resp = await self.handle_req(
                http_request, background_tasks, sub_response, ctx,
                dependency_cache=dependency_cache,
                shared_dependencies_error=shared_dependencies_error,
            )
            ctx.on_raw_response(resp)

        return ctx.raw_response

    async def handle_req(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
        ctx: JsonRpcContext,
        dependency_cache: dict = None,
        shared_dependencies_error: BaseError = None
    ):
        http_request_shadow = RequestShadow(http_request)
        http_request_shadow.scope['path'] = self.path + '/' + ctx.request.method

        for route in self.entrypoint.routes:  # type: MethodRoute
            match, child_scope = route.matches(http_request_shadow.scope)
            if match == Match.FULL:
                ctx.method_route = route
                # http_request is a transport layer and it is common for all JSON-RPC requests in a batch
                return await route.handle_req(
                    http_request_shadow, background_tasks, sub_response, ctx,
                    dependency_cache=dependency_cache,
                    shared_dependencies_error=shared_dependencies_error,
                )
        else:
            raise MethodNotFound()


class Entrypoint(APIRouter):
    method_route_class = MethodRoute
    entrypoint_route_class = EntrypointRoute

    default_errors: Sequence[Type[BaseError]] = [
        InvalidParams, MethodNotFound, ParseError, InvalidRequest, InternalError,
    ]

    def __init__(
        self,
        path: str,
        *,
        name: str = None,
        errors: Sequence[Type[BaseError]] = None,
        dependencies: Sequence[Depends] = None,
        common_dependencies: Sequence[Depends] = None,
        middlewares: Sequence[JsonRpcMiddleware] = None,
        scheduler_factory: Callable[..., Awaitable[aiojobs.Scheduler]] = aiojobs.create_scheduler,
        scheduler_kwargs: dict = None,
        request_class: Type[JsonRpcRequest] = JsonRpcRequest,
        **kwargs,
    ) -> None:
        super().__init__(redirect_slashes=False)
        if errors is None:
            errors = self.default_errors
        self.middlewares = middlewares or []
        self.scheduler_factory = scheduler_factory
        self.scheduler_kwargs = scheduler_kwargs
        self.request_class = request_class
        self.scheduler = None
        self.callee_module = inspect.getmodule(inspect.stack()[1][0]).__name__
        self.entrypoint_route = self.entrypoint_route_class(
            self,
            path,
            name=name,
            errors=errors,
            dependencies=dependencies,
            common_dependencies=common_dependencies,
            request_class=request_class,
            **kwargs,
        )
        self.routes.append(self.entrypoint_route)

    @property
    def common_dependencies(self):
        return self.entrypoint_route.common_dependencies

    async def shutdown(self):
        if self.scheduler is not None:
            await self.scheduler.close()

    async def get_scheduler(self):
        if self.scheduler is not None:
            return self.scheduler
        self.scheduler = await self.scheduler_factory(**(self.scheduler_kwargs or {}))
        return self.scheduler

    async def handle_exception(self, exc):
        raise exc

    async def handle_exception_to_resp(self, exc, log_unhandled_exception=True) -> Tuple[dict, bool]:
        is_unhandled_exception = False
        try:
            resp = await self.handle_exception(exc)
        except BaseError as error:
            resp = error.get_resp()
        except HTTPException:
            raise
        except Exception as exc:
            if log_unhandled_exception:
                logger.exception(str(exc), exc_info=exc)
            resp = InternalError().get_resp()
            is_unhandled_exception = True
        return resp, is_unhandled_exception

    def bind_dependency_overrides_provider(self, value):
        for route in self.routes:
            route.dependency_overrides_provider = value

    async def solve_shared_dependencies(
        self,
        http_request: Request,
        background_tasks: BackgroundTasks,
        sub_response: Response,
    ) -> dict:
        return await self.entrypoint_route.solve_shared_dependencies(
            http_request,
            background_tasks,
            sub_response,
        )

    def add_method_route(
        self,
        func: FunctionType,
        *,
        name: str = None,
        **kwargs,
    ) -> None:
        name = name or func.__name__
        #name = self.entrypoint_route.name + ' - ' + (name or func.__name__)
        route = self.method_route_class(
            self,
            self.entrypoint_route.path + '/' + name,
            func,
            name=name,
            request_class=self.request_class,
            **kwargs,
        )
        self.routes.append(route)

    def method(
        self,
        **kwargs,
    ) -> Callable:
        def decorator(func: FunctionType) -> Callable:
            self.add_method_route(
                func,
                **kwargs,
            )
            return func

        return decorator


class LazyJRPC(FastAPI):
    def openapi(self):
        result = super().openapi()
        for route in self.routes:
            if isinstance(route, (EntrypointRoute, MethodRoute, )):
                route: Union[EntrypointRoute, MethodRoute]
                for media_type in result['paths'][route.path]:
                    result['paths'][route.path][media_type]['responses'].pop('default', None)
        return result

    def bind_entrypoint(self, ep: Entrypoint):
        ep.bind_dependency_overrides_provider(self)
        self.routes.extend(ep.routes)
        self.on_event('shutdown')(ep.shutdown)


import dill
from pydantic import BaseModel, Field, root_validator
from pydantic.types import ByteSize
from typing import Union, Any, Dict, Optional
from lazyops.configs.base import DefaultSettings
from lazyops.types import lazyproperty
from lazyops.libs.sqlcache.constants import DEFAULT_SETTINGS, OPTIMIZED_SETTINGS, DBNAME

def get_eviction_policies(table_name: str):
    return {
        'none': {
            'init': None,
            'get': None,
            'cull': None,
        },
        'least-recently-stored': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS {table_name}_store_time ON'
                f' {table_name} (store_time)'
            ),
            'get': None,
            'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY store_time LIMIT ?',
        },
        'least-recently-used': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS {table_name}_access_time ON'
                f' {table_name} (access_time)'
            ),
            'get': 'access_time = {now}',
            'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY access_time LIMIT ?',
        },
        'least-frequently-used': {
            'init': (
                f'CREATE INDEX IF NOT EXISTS {table_name}_access_count ON'
                f' {table_name} (access_count)'
            ),
            'get': 'access_count = access_count + 1',
            'cull': 'SELECT {fields} FROM ' + table_name + ' ORDER BY access_count LIMIT ?',
        },
    }


class SqlCacheSettings(DefaultSettings):
    """
    Settings for the SqlCache.
    """

    class Config(DefaultSettings.Config):
        env_prefix = 'SQLCACHE_'
        case_sensitive = False
    

class SqlCacheConfig(BaseModel):

    table_name: str = 'sqlcache'
    db_name: str = DBNAME
    statistics: Union[int, bool] = False
    tag_index: Union[int, bool] = False
    eviction_policy: str = 'least-recently-stored'
    size_limit: ByteSize = Field(default = OPTIMIZED_SETTINGS['standard']['size_limit'])
    cull_limit: int = 10
    sqlite_auto_vacuum: int = 1  # FULL
    cache_size: int = 2 ** 13  # 8,192 pages
    sqlite_journal_mode: str = 'wal'
    mmap_size: ByteSize = Field(default = OPTIMIZED_SETTINGS['standard']['mmap_size']) # 2**26  # 64mb
    sqlite_synchronous: int = 1  # NORMAL
    disk_min_file_size: ByteSize = Field(default = OPTIMIZED_SETTINGS['standard']['min_file_size']) # 2**15  # 32kb
    disk_pickle_protocol: int = 4
    compression_level: int = Field(default = OPTIMIZED_SETTINGS['standard']['compression_level'])
    dataset_mode: bool = False # if enabled, will start Index at 0 rather than 500 trill

    @root_validator(pre = True)
    def validate_config(cls, values: Dict) -> Dict:
        """
        Ensures that the configuration is valid.
        """
        if 'statistics' in values and isinstance(values['statistics'], bool):
            values['statistics'] = int(values['statistics'])
        if 'tag_index' in values and isinstance(values['tag_index'], bool):
            values['tag_index'] = int(values['tag_index'])
        return values

    @classmethod
    def from_optimized(cls, table_name: str, optim: Optional[str] = 'standard', config: Optional[Dict[str, Any]] = None) -> 'SqlCacheConfig':
        """
        Allows for the creation of a SqlCacheConfig object from a pre-defined
        optimized configuration. The `config` parameter allows for the
        """

        assert optim in OPTIMIZED_SETTINGS, f'Invalid optimization: {optim}'
        base_config = OPTIMIZED_SETTINGS[optim].copy()
        if config: base_config.update(config)
        return cls(table_name = table_name, **base_config)

    @property
    def start_index_n(self):
        return 0 if self.dataset_mode else 500000000000000

    @property
    def sql_settings(self) -> Dict[str, Any]:
        return {
            'statistics': self.statistics,  # False
            'tag_index': self.tag_index,  # False
            'eviction_policy': self.eviction_policy,
            'size_limit': self.size_limit,
            'cull_limit': self.cull_limit,
            'sqlite_auto_vacuum': self.sqlite_auto_vacuum, 
            'sqlite_cache_size': self.cache_size,
            'sqlite_journal_mode': self.sqlite_journal_mode,
            'sqlite_mmap_size': self.mmap_size,
            'sqlite_synchronous': self.sqlite_synchronous,  # NORMAL
            'disk_min_file_size': self.disk_min_file_size,
            'disk_pickle_protocol': self.disk_pickle_protocol,
        }

    @property
    def eviction_policy_config(self):
        return get_eviction_policies(self.table_name)[self.eviction_policy]
    




from lazyops.libs.sqlcache.constants import ENOVAL
from typing import Dict, Tuple


def full_name(func):
    "Return full name of `func` by adding the module and function name."
    return f'{func.__module__}.{func.__qualname__}'


def args_to_key(base: Tuple, args: Tuple, kwargs: Dict, typed: bool) -> Tuple:
    """Create cache key out of function arguments.
    :param tuple base: base of key
    :param tuple args: function arguments
    :param dict kwargs: function keyword arguments
    :param bool typed: include types in cache key
    :return: cache key tuple
    """
    key = base + args

    if kwargs:
        key += (ENOVAL,)
        sorted_items = sorted(kwargs.items())
        for item in sorted_items:
            key += item

    if typed:
        key += tuple(type(arg) for arg in args)
        if kwargs:
            key += tuple(type(value) for _, value in sorted_items)

    return key




"""
Import Handler for `torch`
"""
import sys
import typing
from lazyops.utils.imports import resolve_missing_custom, require_missing_wrapper

try:
    import torch
    _torch_available = True
except ImportError:
    torch = object
    _torch_available = False

def resolve_torch(
    required: bool = False,
    version: typing.Optional[str] = None,
    cuda: typing.Optional[str] = None,
    rocm: typing.Optional[str] = None,
):
    """
    Ensures that `torch` is available

    :arg cuda: If `True`, will attempt to import `torch` with CUDA support. 
    If str will attempt to import `torch` with the specified CUDA version.
    :type cuda: `str` or `bool`
    """
    global torch, _torch_available
    if not _torch_available:
        pkg = 'torch'
        is_win = sys.platform.startswith('win')
        if version is not None:
            pkg += f'=={version}'
            if cuda is not None:
                pkg += f'+cu{cuda}'
            elif rocm is not None:
                pkg += f'+rocm{rocm}'
            elif is_win:
                pkg += '+cpu'
            
        pkg += ' --extra-index-url https://download.pytorch.org/whl/torch'
        resolve_missing_custom(pkg, required = required)
        import torch
        _torch_available = True
        globals()['torch'] = torch


def require_torch(
    required: bool = False,
    version: typing.Optional[str] = None,
    cuda: typing.Union[str, bool] = False,
    rocm: typing.Optional[str] = None,
):
    """
    Wrapper for `resolve_torch` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_torch, 
            func = func, 
            required = required,
            version = version,
            cuda = cuda,
            rocm = rocm,
        )
    return decorator


# Utilities

import dill
import random
import signal
import anyio
from aiocache import cached
from aiocache.serializers import PickleSerializer, BaseSerializer
from anyio.abc import CancelScope
from typing import List, Union, Callable, Any
from lazyops.utils.logs import logger

class InfiniteBackoffsWithJitter:
    def __iter__(self):
        while True:
            yield 10 + random.randint(-5, +5)

class DillSerializerObject(BaseSerializer):
    """
    Transform data to bytes using Dill.dumps and Dill.loads to retrieve it back.
    """

    DEFAULT_ENCODING = None
    PROTOCOL = dill.HIGHEST_PROTOCOL

    def dumps(self, value):
        """
        Serialize the received value using ``Dill.dumps``.

        :param value: obj
        :returns: bytes
        """
        return dill.dumps(value, protocol = self.PROTOCOL)

    def loads(self, value):
        """
        Deserialize value using ``Dill.loads``.

        :param value: bytes
        :returns: obj
        """
        return None if value is None else dill.loads(value)

DillSerializer: DillSerializerObject = DillSerializerObject()


"""
Utility Methods

- include signal handler to handle SIGINT
borrowed from: https://github.com/gexcom/pgsync-controller/blob/main/core/main.py
"""


class SignalHandler:

    _watch: List[signal.Signals] = [signal.SIGINT, signal.SIGTERM, signal.SIGKILL, signal.SIGSTOP, signal.SIGQUIT] # Handle SigInt, SigTerm, SigKill, SigStop, SigQuit
    _exit_funcs: List[Callable[[Any, Any], Any]] = []

    @classmethod
    async def monitor(cls, exit_funcs: List[Callable] = [], **kwargs):
        """
        Monitor SIGINT, SIGTERM, SIGKILL, SIGSTOP, SIGQUIT
        When received, call runs all exit_funcs and cls._exit_funcs

        Don't know what to do after this part though.
        """
        
        _stop_event = anyio.Event()
        interrupted = False
        interrupt_signal = None
        
        async def _signal_handler(scope: CancelScope) -> None:
            nonlocal interrupted
            nonlocal interrupt_signal
            with anyio.open_signal_receiver(*cls._watch) as signals:
                async for s in signals:
                    interrupted = True
                    interrupt_signal = s
                    _stop_event.set()
                    scope.cancel()
                    break

        CancelledError = anyio.get_cancelled_exc_class()

        while True:
            logger.info('Starting Signal Handler')
            async with anyio.create_task_group() as tg:
                tg.start_soon(_signal_handler, tg.cancel_scope)
                


    def __init__(self, sig_num, handler):
        self.sig_num = sig_num
        self.handler = handler

    def __enter__(self):
        self.old_handler = signal.getsignal(self.sig_num)
        signal.signal(self.sig_num, self.handler)

    def __exit__(self, *args):
        signal.signal(self.sig_num, self.old_handler)


from __future__ import annotations

"""
Stateful Registry
"""

from typing import Dict, TypeVar, Optional, Type, Union, TYPE_CHECKING
from lazyops.utils.lazy import lazy_import
from lazyops.utils.logs import logger

ClientT = TypeVar('ClientT')
ClientTypeT = Type[ClientT]

_registered_clients: Dict[str, Union[ClientT, ClientTypeT]] = {}
_client_registry_mapping: Dict[str, str] = None # A mapping of client names to module paths

def update_client_registry_mapping(
    mapping: Dict[str, str],
):
    """
    Updates the client registry mapping
    """
    global _client_registry_mapping
    if _client_registry_mapping is None:
        _client_registry_mapping = {}
    _client_registry_mapping.update(mapping)


def get_global_client(
    name: str,
) -> ClientTypeT:
    """
    Global Clients do not require initialization and are types
    """
    global _registered_clients
    if not _client_registry_mapping:
        raise ValueError('Client Registry Mapping not set')
    if name not in _registered_clients:
        if name not in _client_registry_mapping:
            raise ValueError(f'Client {name} not found in client registry mapping')
        _registered_clients[name] = lazy_import(_client_registry_mapping[name])
    return _registered_clients[name]


def get_client(
    name: str, 
    **kwargs,
) -> ClientT:
    """
    Clients require initialization
    """
    global _registered_clients
    if not _client_registry_mapping:
        raise ValueError('Client Registry Mapping not set')
    if name not in _registered_clients:
        if name not in _client_registry_mapping:
            raise ValueError(f'Client {name} not found in client registry mapping')
        client = lazy_import(_client_registry_mapping[name])
        _registered_clients[name] = client(**kwargs)
    return _registered_clients[name]

def register_client(
    client: Union[ClientT, ClientTypeT],
    name: str,
    verbose: Optional[bool] = False,
    **kwargs,
) -> bool:
    """
    Registers a client
    """
    global _registered_clients
    if name not in _registered_clients:
        _registered_clients[name] = client
        if verbose: logger.info(f'Client {name}', colored = True, prefix = "|g|Registered|e|")
        return True
    return False




import datetime
import contextlib
from sqlalchemy import func
from sqlalchemy import delete as sqlalchemy_delete
from sqlalchemy import update as sqlalchemy_update
from sqlalchemy import exists as sqlalchemy_exists
from sqlalchemy import insert as sqlalchemy_insert
from sqlalchemy.future import select
from sqlalchemy.sql.expression import Select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import selectinload, joinedload, immediateload
from sqlalchemy import Column, Integer, DateTime, String, Text, ForeignKey, Boolean, Identity, Enum
from sqlalchemy.dialects.postgresql import insert as postgres_insert
from typing import Any, Generator, AsyncGenerator, Iterable, Optional, Union, Type, Dict, cast, TYPE_CHECKING, List, Tuple, TypeVar, Callable
from lazyops.utils import create_unique_id, create_timestamp
from lazyops.utils.logs import logger
from lazyops.types import lazyproperty
from lazyops.libs.psqldb.base import Base, PostgresDB, AsyncSession, Session
from lazyops.libs.psqldb.utils import SQLJson, get_pydantic_model, object_serializer
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

ModelType = TypeVar("ModelType", bound = Base)

_attr_func_map = {
    "selectinload": selectinload,
    "joinedload": joinedload,
    "immediateload": immediateload,
}

def get_attr_func(
    attr: Union[str, Callable] = None,
):
    if attr and isinstance(attr, str):
        return _attr_func_map.get(attr, selectinload)
    return attr or selectinload

BaseSQLModelT = TypeVar('BaseSQLModelT', bound = 'BaseSQLModel')

class BaseSQLModel(Base):
    """
    Abstract base class for all SQL models
    """
    __abstract__ = True
    __allow_unmapped__ = True
    
    id: str = Column(Text, default = create_unique_id, primary_key = True, index = True)
    created_at: datetime.datetime = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now())
    updated_at: datetime.datetime = Column(DateTime(timezone=True), default = create_timestamp, server_default = func.now(), onupdate = create_timestamp)

    @lazyproperty
    def pydantic_model(self) -> Type[BaseModel]:
        """
        Return the Pydantic model for this ORM model.
        """
        return get_pydantic_model(self)

    @classmethod
    def _build_query(
        cls, 
        query: Optional[Select] = None, 
        load_attrs: Optional[List[str]] = None,
        load_attr_method: Optional[Union[str, Callable]] = None,
        **kwargs
    ) -> Select:
        """
        Build a query
        """
        query = query if query is not None else select(cls)
        query = query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])
        if load_attrs:
            load_attr_method = get_attr_func(load_attr_method)
            for attr in load_attrs:
                query = query.options(load_attr_method(getattr(cls, attr)))
        return query

    def dict(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        safe_encode: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        data = self.pydantic_model.dict(
            exclude = exclude, include = include, **kwargs
        )
        if safe_encode: data = {key: object_serializer(value) for key, value in data.items()}
        return data
    
    def json(
        self,
        exclude: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude_none: Optional[bool] = False,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Return a dictionary representation of the model.
        """
        return self.pydantic_model.json(
            exclude = exclude, include = include, exclude_none = exclude_none, **kwargs
        )

    @classmethod
    def _filter(cls, query: Optional[Select] = None, **kwargs) -> Select:
        """
        Build a filter query
        """
        query = query if query is not None else select(cls)
        return query.where(*[getattr(cls, key) == value for key, value in kwargs.items()])

    def _filter_update_data(self, **kwargs) -> Optional[Dict[str, Any]]:
        """
        Filter update data
        """
        data = {}
        for field, value in kwargs.items():
            if not hasattr(self, field): continue
            with contextlib.suppress(ValueError):
                if hasattr(value, 'all') and hasattr(getattr(self, field), 'all') \
                    and getattr(self, field).all() == value.all():
                    continue
                if getattr(self, field) == value: 
                    continue
            data[field] = value
        return data or None

    @classmethod
    def _handle_exception(
        cls, 
        msg: Optional[str] = None, 
        error: Optional[Exception] = None, 
        verbose: Optional[bool] = False
    ):
        """
        Handle exception
        """
        msg = msg or f"{cls.__name__} not found"
        if verbose: logger.trace(msg, error = error)
        raise error or HTTPException(status_code = 404, detail = msg)

    @classmethod
    def session(cls, session: Optional[Session] = None, **kwargs):
        """
        Get session
        """
        return PostgresDB.session(session = session, **kwargs)
    
    @classmethod
    def async_session(cls, session: Optional[AsyncSession] = None, **kwargs):
        """
        Get async session
        """
        return PostgresDB.async_session(session = session, **kwargs)
    
    def cast_to(self, model: ModelType) -> ModelType:
        """
        Cast to another model
        """
        return cast(model, self)

    @contextlib.contextmanager
    @classmethod
    def safe_ctx(
        cls,
        func: str,
        default: Optional[Any] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ):
        """
        Session context manager to handle exceptions
        """
        try:
            yield
        except Exception as e:
            if _raise_exceptions:
                cls._handle_exception(error = e, verbose = _verbose)
            elif _verbose:
                logger.trace(f'Error {func} for {cls.__name__}', error = e)
            return default
        finally:
            pass


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.dict()})>"
        

SyncSQLModelT = TypeVar('SyncSQLModelT', bound = 'SyncSQLModel')

class SyncSQLModel(BaseSQLModel):
    """
    Synchronous SQL model
    """
    __abstract__ = True
    __allow_unmapped__ = True

    @classmethod
    def create(
        cls,
        session: Optional[Session] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> SyncSQLModelT:
        """
        Create a record
        """
        with cls.session(session = session) as db_sess:
            try:
                record = cls(**kwargs)
                db_sess.add(record)
                db_sess.commit()
                db_sess.refresh(record)
                return record
            except Exception as e:
                if _raise_exceptions:
                    cls._handle_exception(error = e, verbose = _verbose)
                elif _verbose:
                    logger.trace(f'Error creating {cls.__name__}', error = e)
                return None

    @classmethod
    def get(
        cls, 
        session: Optional[Session] = None,
        _load_attrs: Optional[List[str]] = None,
        _load_attr_method: Optional[Union[str, Callable]] = None,
        _readonly: Optional[bool] = False,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> SyncSQLModelT:
        """
        Get a record
        """
        with cls.session(session = session, ro = _readonly) as db_sess:
            query = cls._build_query(load_attrs = _load_attrs, load_attr_method = _load_attr_method, **kwargs)
            result = (db_sess.execute(query)).scalar_one_or_none()
            if not result and _raise_exceptions:
                cls._handle_exception(error = NoResultFound(), verbose = _verbose)
            return result

    @classmethod
    def get_all(
        cls,
        session: Optional[Session] = None,
        _skip: Optional[int] = None, 
        _limit: Optional[int] = None,
        _load_attrs: Optional[List[str]] = None,
        _load_attr_method: Optional[Union[str, Callable]] = None,
        _readonly: Optional[bool] = False,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs,
    ) -> List[SyncSQLModelT]:
        """
        Get all records
        """
        with cls.session(session = session, ro = _readonly) as db_sess:
            query = cls._build_query(load_attrs = _load_attrs, load_attr_method = _load_attr_method, **kwargs)
            if _skip is not None: query = query.offset(_skip)
            if _limit is not None: query = query.limit(_limit)
            results = (db_sess.scalars(query)).all()
            if not results and _raise_exceptions:
                cls._handle_exception(error = NoResultFound(), verbose = _verbose)
            return results
    
    @classmethod
    def get_first(
        cls, 
        session: Optional[Session] = None,
        _load_attrs: Optional[List[str]] = None,
        _load_attr_method: Optional[Union[str, Callable]] = None,
        _readonly: Optional[bool] = False,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> Optional[SyncSQLModelT]:
        """
        Return the first result of a query.
        """
        with cls.session(session = session, ro = _readonly) as db_sess:
            query = cls._build_query(load_attrs = _load_attrs, load_attr_method = _load_attr_method, **kwargs)
            result = db_sess.scalars(query).first()
            if not result and _raise_exceptions:
                cls._handle_exception(error = NoResultFound(), verbose = _verbose)
            return result
    
    @classmethod
    def get_batch(
        cls,
        session: Optional[Session] = None,
        _skip: Optional[int] = None, 
        _limit: Optional[int] = None,
        **kwargs
    ) -> Optional[List[SyncSQLModelT]]:
        """
        Returns the existing records
        """
        with cls.session(session = session) as db_sess:
            query = select(cls)
            for key, value in kwargs.items():
                if isinstance(value, (list, tuple, set)):
                    query = query.filter(getattr(cls, key).in_(value))
                else:
                    query = query.filter(getattr(cls, key) == value)
            if _skip is not None: query = query.offset(_skip)
            if _limit is not None: query = query.limit(_limit)
            return db_sess.execute(query).scalars().all()
    
    @classmethod
    def delete(
        cls, 
        session: Optional[Session] = None, 
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> Optional[SyncSQLModelT]:
        """
        Delete a record
        """
        with cls.session(session = session) as db_sess:
            obj = cls.get(session = db_sess, _raise_exceptions = _raise_exceptions, _verbose = _verbose, **kwargs)
            if obj is not None:
                db_sess.delete(obj)
                db_sess.commit()
            return obj
        
    @classmethod
    def delete_batch(
        cls,
        session: Optional[Session] = None,
        _skip: Optional[int] = None, 
        _limit: Optional[int] = None,
        **kwargs
    ) -> int:
        """
        Delete batch of records
        """
        with cls.session(session = session) as db_sess:
            results = cls.get_batch(
                session = db_sess,
                _skip = _skip,
                _limit = _limit,
                **kwargs
            )
            if results:
                with db_sess.begin():
                    for result in results:
                        db_sess.delete(result)
                db_sess.commit()
            return len(results)
    
    @classmethod
    def delete_all(
        cls,
        session: Optional[Session] = None,
        _raise_exceptions: Optional[bool] = False,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> int:
        """
        Delete all records
        """
        with cls.session(session = session) as db_sess:
            # Get the number of rows that will be deleted
            num_records = cls.count(session = db_sess, _raise_exceptions = _raise_exceptions, _verbose = _verbose, **kwargs)
            query = sqlalchemy_delete(cls)
            with cls.safe_ctx('delete all records', default = 0, _raise_exceptions = _raise_exceptions, _verbose = _verbose):
                db_sess.execute(query)
                db_sess.commit()
                return num_records


    @classmethod
    def count(
        cls,
        session: Optional[Session] = None,
        _readonly: Optional[bool] = True,
        _raise_exceptions: Optional[bool] = False,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> int:
        """
        Return the number of records
        """
        with cls.session(session = session, ro = _readonly) as db_sess:
            query = cls._filter(**kwargs) if kwargs else select(func.count(cls.id))
            with cls.safe_ctx('retrieving count', default = 0, _raise_exceptions = _raise_exceptions, _verbose = _verbose):
                return db_sess.execute(query).scalar()
            
    @classmethod
    def exists(
        cls,
        session: Optional[Session] = None,
        _readonly: Optional[bool] = True,
        _raise_exceptions: Optional[bool] = False,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> bool:
        """
        Return True if the record exists
        """
        with cls.session(session = session, ro = _readonly) as db_sess:
            query = cls._filter(sqlalchemy_exists(), **kwargs).select()
            try:
                return db_sess.execute(query).scalar()
            except Exception as e:
                if _raise_exceptions:
                    cls._handle_exception(error = e, verbose = _verbose)
                elif _verbose:
                    logger.trace(f'Error retrieving exists for {cls.__name__}', error = e)
                return False

    def update(
        self,
        session: Optional[Session] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> SyncSQLModelT:
        """
        Update a record
        """
        with self.session(session = session) as db_sess:
            query = sqlalchemy_update(self.__class__).where(
                self.__class__.id == self.id).values(
                **kwargs
            )
            try:
                db_sess.execute(query)
                db_sess.commit()
            except Exception as e:
                if _raise_exceptions:
                    self._handle_exception(error = e, verbose = _verbose)
                elif _verbose:
                    logger.trace(f'Error updating {self.__class__.__name__}', error = e)
            return self

    def update_inplace(
        self,
        session: Optional[Session] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> SyncSQLModelT:
        """
        Update a record in place
        """
        with self.session(session = session) as db_sess:
            for key, value in kwargs.items():
                setattr(self, key, value)
            try:
                local_object = db_sess.merge(self)
                db_sess.add(local_object)
                db_sess.commit()
                db_sess.flush()
                self = local_object
            except Exception as e:
                if _raise_exceptions:
                    self._handle_exception(error = e, verbose = _verbose)
                elif _verbose:
                    logger.trace(f'Error updating {self.__class__.__name__}', error = e)
            return self
    
    @classmethod
    def upsert(
        cls,
        session: Optional[Session] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> SyncSQLModelT:
        """
        Upsert a record
        """
        with cls.session(session = session) as db_sess:
            query = sqlalchemy_insert(cls).values(**kwargs)
            query = query.on_conflict_do_update(
                index_elements = [cls.id],
                set_ = kwargs
            )
            try:
                db_sess.execute(query)
                db_sess.commit()
            except Exception as e:
                if _raise_exceptions:
                    cls._handle_exception(error = e, verbose = _verbose)
                elif _verbose:
                    logger.trace(f'Error upserting {cls.__name__}', error = e)
            return cls.get(session = db_sess, _raise_exceptions = _raise_exceptions, _verbose = _verbose, **kwargs)

    @classmethod    
    def get_or_create(
        cls,
        session: Optional[Session] = None,
        _filterby: Optional[Iterable[str]] = None,
        _raise_exceptions: Optional[bool] = True,
        _verbose: Optional[bool] = False,
        **kwargs
    ) -> Tuple[SyncSQLModelT, bool]:
        """
        Create a record if it doesn't exist

        Returns:
            Tuple[SyncSQLModelT, bool]: The record and a boolean indicating if it was created
        """
        with cls.session(session = session) as db_sess:
            _filterby = [list(kwargs.keys())[0]] if _filterby is None else _filterby
            filterby = {key: kwargs.get(key) for key in _filterby}
            result = cls.get(
                session = session,
                _raise_exceptions = False, 
                _verbose = False, 
                **filterby, 
            )
            if result is not None: return result, False
            return cls.create(session = session, _raise_exceptions = _raise_exceptions, _verbose = _verbose, **kwargs), True
    
    @classmethod
    def get_or_none(
        cls,
        session: Optional[Session] = None,
        **kwargs
    ) -> Optional[SyncSQLModelT]:
        """
        Get a record or None if it doesn't exist
        """
        return cls.get(session = session, _raise_exceptions = False, _verbose = False, **kwargs)
    




"""
Import Handler for backoff
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import backoff
    _backoff_available = True
except ImportError:
    backoff = object
    _backoff_available = False

def resolve_backoff(
    required: bool = False,
):
    """
    Ensures that `backoff` is available
    """
    global backoff, _backoff_available
    if not _backoff_available:
        resolve_missing('backoff', required = required)
        import backoff
        _backoff_available = True
        globals()['backoff'] = backoff


def require_backoff(
    required: bool = False,
):
    """
    Wrapper for `resolve_backoff` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_backoff, 
            func = func, 
            required = required
        )
    return decorator


from enum import Enum

class Privilege(str, Enum):
    """
    Enumeration of all possible privileges, see official documentation `here <https://www.postgresql.org/docs/current/ddl-priv.html#PRIVILEGE-ABBREVS-TABLE>`_.
    """

    ALL_PRIVILEGES = "ALL PRIVILEGES"
    SELECT = "SELECT"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    TRUNCATE = "TRUNCATE"
    REFERENCES = "REFERENCES"
    TRIGGER = "TRIGGER"
    USAGE = "USAGE"
    CREATE = "CREATE"
    CONNECT = "CONNECT"
    TEMPORARY = "TEMPORARY"
    EXECUTE = "EXECUTE"
    ALTER_SYSTEM = "ALTER SYSTEM"



"""
LZ4 Compress
"""

from .base import BaseCompress
from typing import Union

try:
    import lz4.frame
    _lz4_available = True
except ImportError:
    _lz4_available = False

def ensure_lz4_available():
    """
    Ensure LZ4 is available
    """
    global _lz4_available, lz4
    if _lz4_available is False: 
        from lazyops.utils.imports import resolve_missing
        resolve_missing('lz4', required = True)
        import lz4.frame
        _lz4_available = True
        globals()['lz4'] = lz4


class Lz4Compress(BaseCompress):


    @classmethod
    def compress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        LZ4 Compress
        """
        ensure_lz4_available()
        if isinstance(data, str): data = data.encode(encoding)
        return lz4.frame.compress(data)

    @classmethod
    def decompress(cls, data: Union[str, bytes], encoding: str = 'utf-8', **kwargs) -> bytes:
        """
        LZ4 Decompress
        """
        ensure_lz4_available()
        if isinstance(data, str): data = data.encode(encoding = encoding)
        return lz4.frame.decompress(data)





import atexit
from .types import GlobalContext, GracefulKiller
from .workers import spawn_new_worker, stop_worker, terminate_worker


def register_exit():
    """
    Registers the exit function
    """
    atexit.register(GlobalContext.end_all_processes)


def run_until_complete():
    """
    Runs the event loop until complete
    """
    import time
    from lazyops.utils import logger
    while True:
        try: time.sleep(5.0)
        except KeyboardInterrupt:
            logger.warning("Keyboard Interrupt")
            break
        except Exception as e:
            logger.error(f"Error: {e}")
            break
    
    GlobalContext.end_all_processes()


async def arun_until_complete(
    termination_file: str = None,
):
    """
    Runs the event loop until complete
    """
    import os
    import asyncio
    import pathlib
    from lazyops.utils import logger

    if termination_file is None:
        termination_file = os.getenv('WORKER_TERMINATION_FILE')

    watch = GracefulKiller()
    tmp_kill_file = pathlib.Path(termination_file) if termination_file is not None else None
    while not watch.kill_now:
        try: 
            await asyncio.sleep(1.0)
            if tmp_kill_file is not None and tmp_kill_file.exists():
                logger.warning(f"Found termination file: {tmp_kill_file}")
                break
        except KeyboardInterrupt:
            logger.warning("Keyboard Interrupt")
            break
        except Exception as e:
            logger.error(f"Error: {e}")
            break
    
    GlobalContext.end_all_processes()
    await GlobalContext.aclose_processes()



"""
Import Handler for `kubernetes`, `kubernetes_asyncio`, `kopf`
"""

try:

    import kubernetes.client as SyncClient
    import kubernetes.config as SyncConfig
    import kubernetes.stream.ws_client as SyncStream
    import kubernetes.stream.stream as SyncStreamFunc
    
    import kubernetes.watch as SyncWatch
    import kubernetes.utils as SyncUtils
    import kubernetes.client.models as SyncType

    _k8_available = True

except ImportError:
    SyncClient = object
    SyncConfig = object
    SyncStream = object
    SyncStreamFunc = object

    SyncWatch = object
    SyncUtils = object
    SyncType = object

    _k8_available = False

try:
    import kubernetes_asyncio.client as AsyncClient
    import kubernetes_asyncio.config as AsyncConfig
    import kubernetes_asyncio.stream as AsyncStream
    import kubernetes_asyncio.watch as AsyncWatch
    import kubernetes_asyncio.utils as AsyncUtils
    import kubernetes_asyncio.client.models as AsyncType

    AsyncWSClient = AsyncStream.WsApiClient

    _async_k8_available = True

except ImportError:
    AsyncClient = object
    AsyncConfig = object
    AsyncStream = object
    AsyncWatch = object
    AsyncUtils = object
    AsyncType = object
    AsyncWSClient = object

    _async_k8_available = False

try:
    import kopf
    _kopf_available = True

except ImportError:
    kopf = object
    _kopf_available = False

try:
    import aiocache
    _aiocache_available = True

except ImportError:
    aiocache = object
    _aiocache_available = False

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

def resolve_sync_k8s(
    required: bool = True,
):
    """
    Ensures that `kubernetes` is availableable
    """
    global _k8_available
    global SyncClient, SyncConfig, SyncStream, SyncWatch, SyncUtils, SyncType, SyncStreamFunc

    if not _k8_available:
        resolve_missing('kubernetes', required = required)
        import kubernetes.client as SyncClient
        import kubernetes.config as SyncConfig
        import kubernetes.stream.ws_client as SyncStream
        import kubernetes.stream.stream as SyncStreamFunc
        import kubernetes.watch as SyncWatch
        import kubernetes.utils as SyncUtils
        import kubernetes.client.models as SyncType        

        _k8_available = True

def resolve_async_k8s(
    required: bool = True,
):
    """
    Ensures that `kubernetes_asyncio` is availableable
    """
    global _async_k8_available
    global AsyncClient, AsyncConfig, AsyncStream, AsyncWatch, AsyncUtils, AsyncType, AsyncWSClient
    
    if not _async_k8_available:
        resolve_missing('kubernetes_asyncio', required = required)
        import kubernetes_asyncio.client as AsyncClient
        import kubernetes_asyncio.config as AsyncConfig
        import kubernetes_asyncio.stream as AsyncStream
        import kubernetes_asyncio.watch as AsyncWatch
        import kubernetes_asyncio.utils as AsyncUtils
        import kubernetes_asyncio.client.models as AsyncType
        AsyncWSClient = AsyncStream.WsApiClient
        _async_k8_available = True

def resolve_aiocache(
    required: bool = True,
):
    """
    Ensures that `aiocache` is availableable
    """
    global _aiocache_available
    global aiocache

    if not _aiocache_available:
        resolve_missing('aiocache', required = required)
        import aiocache
        _aiocache_available = True
        globals()['aiocache'] = aiocache

def resolve_kopf(
    required: bool = True,
):
    """
    Ensures that `kopf`
    """
    global _kopf_available
    global kopf

    if not _kopf_available:
        resolve_missing('kopf', required = required)
        import kopf
        _kopf_available = True
        globals()['kopf'] = kopf

def resolve_k8s(
    is_sync: bool = False,
    is_async: bool = True,
    is_operator: bool = False,
    required: bool = True,
):
    """
    Ensures that `kubernetes`, `kubernetes_asyncio`, `kopf` are availableable
    """
    if is_operator:
        resolve_kopf(required = required)
    if is_async:
        resolve_async_k8s(required = required)
        resolve_aiocache(required = required)
    if is_sync or not is_async:
        resolve_sync_k8s(required = required)


def require_k8s(
    is_sync: bool = False,
    is_async: bool = True,
    is_operator: bool = False,
    required: bool = True,
):
    """
    Wrapper for `resolve_k8s` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_k8s, 
            func = func, 
            is_sync = is_sync,
            is_async = is_async, 
            is_operator = is_operator,
            required = required
        )
    return decorator



import math
import fileio
import collections
import platform
import multiprocessing as multiproc
import random
from functools import reduce
from itertools import chain, count, islice, takewhile

from typing import List, Optional, Dict

_CPU_CORES = multiproc.cpu_count()
_MAX_PROCS = math.ceil(_CPU_CORES * 1.5)
_MAX_THREADS = math.ceil(_CPU_CORES * 2)

try:
    _SERIALIZER = fileio.src._pickler
except:
    _SERIALIZER = fileio.core.libs.PICKLE_FUNC
    
_PROTOCOL = _SERIALIZER.HIGHEST_PROTOCOL



def is_primitive(val):
    """
    :param val: value to check
    :return: True if value is a primitive, else False
    """
    return isinstance(val, (str, bool, float, complex, bytes, int))


def is_namedtuple(val):
    """
    Use Duck Typing to check if val is a named tuple. Checks that val is of type tuple and contains
    the attribute _fields which is defined for named tuples.
    :param val: value to check type of
    :return: True if val is a namedtuple
    """
    val_type = type(val)
    bases = val_type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(val_type, "_fields", None)
    return all(isinstance(n, str) for n in fields)


def identity(arg):
    """
    Function which returns the argument. Used as a default lambda function.
    :param arg: object to take identity of
    :return: return arg
    """
    return arg


def is_iterable(val):
    """
    Check if val is not a list, but is a collections.Iterable type. This is used to determine
    when list() should be called on val
    :param val: value to check
    :return: True if it is not a list, but is a collections.Iterable
    """
    if isinstance(val, list):
        return False
    return isinstance(val, collections.abc.Iterable)


def is_tabulatable(val):
    if is_primitive(val):
        return False
    if is_iterable(val) or is_namedtuple(val) or isinstance(val, list):
        return True
    return False


def split_every(parts, iterable):
    """
    Split an iterable into parts of length parts
    :param iterable: iterable to split
    :param parts: number of chunks
    :return: return the iterable split in parts
    """
    return takewhile(bool, (list(islice(iterable, parts)) for _ in count()))


def unpack(packed):
    """
    Unpack the function and args then apply the function to the arguments and return result
    :param packed: input packed tuple of (func, args)
    :return: result of applying packed function on packed args
    """
    func, args = _SERIALIZER.loads(packed)
    result = func(*args)
    if isinstance(result, collections.abc.Iterable):
        return list(result)
    return None


def pack(func, args):
    """
    Pack a function and the args it should be applied to
    :param func: Function to apply
    :param args: Args to evaluate with
    :return: Packed (func, args) tuple
    """
    return _SERIALIZER.dumps((func, args), _PROTOCOL)


def parallelize(func, result, processes=None, partition_size=None):
    """
    Creates an iterable which is lazily computed in parallel from applying func on result
    :param func: Function to apply
    :param result: Data to apply to
    :param processes: Number of processes to use in parallel
    :param partition_size: Size of partitions for each parallel process
    :return: Iterable of applying func on result
    """
    parallel_iter = lazy_parallelize(
        func, result, processes=processes, partition_size=partition_size
    )
    return chain.from_iterable(parallel_iter)


def lazy_parallelize(func, result, processes=None, partition_size=None):
    """
    Lazily computes an iterable in parallel, and returns them in pool chunks
    :param func: Function to apply
    :param result: Data to apply to
    :param processes: Number of processes to use in parallel
    :param partition_size: Size of partitions for each parallel process
    :return: Iterable of chunks where each chunk as func applied to it
    """
    if processes is None or processes < 1:
        processes = _MAX_PROCS
    else:
        processes = min(processes, _CPU_CORES)
    partition_size = partition_size or compute_partition_size(result, processes)
    pool = multiproc.Pool(processes=processes)
    partitions = split_every(partition_size, iter(result))
    packed_partitions = (pack(func, (partition,)) for partition in partitions)
    yield from pool.imap(unpack, packed_partitions)
    pool.terminate()


def compute_partition_size(result, processes):
    """
    Attempts to compute the partition size to evenly distribute work across processes. Defaults to
    1 if the length of result cannot be determined.
    :param result: Result to compute on
    :param processes: Number of processes to use
    :return: Best partition size
    """
    try:
        return max(math.ceil(len(result) / processes), 1)
    except TypeError:
        return 1


def compose(*functions):
    """
    Compose all the function arguments together
    :param functions: Functions to compose
    :return: Single composed function
    """
    return reduce(lambda f, g: lambda x: f(g(x)), functions, lambda x: x)

if __name__ == '__main__':
    if platform.system() == "Darwin":
        n = multiproc.get_context()
        print(n)
        multiproc.set_start_method('spawn')


from __future__ import annotations

"""
ThreadPool Implementation

- v3: Transform into a singleton proxy object class
"""

import os
import abc
import sys
import anyio
import inspect
import asyncio
import functools
import subprocess
import contextvars
import anyio.from_thread
from concurrent import futures
from anyio._core._eventloop import threadlocals
from lazyops.libs.proxyobj import ProxyObject
from typing import Callable, Coroutine, Any, Union, List, Set, Tuple, TypeVar, Optional, Generator, Awaitable, Iterable, AsyncGenerator, Dict

if sys.version_info < (3, 10):
    # Add aiter and anext to asyncio
    def aiter(it: AsyncIterable) -> Any:
        return it.__aiter__()
    
    def anext(it: AsyncIterator) -> Any:
        return it.__anext__()
    

_concurrency_limit: Optional[int] = None

def set_concurrency_limit(
    limit: Optional[int] = None
):
    """
    Set the concurrency limit
    """
    global _concurrency_limit
    if limit is None: limit = os.cpu_count() * 4
    _concurrency_limit = limit

def get_concurrency_limit() -> Optional[int]:
    """
    Get the concurrency limit
    """
    global _concurrency_limit
    if _concurrency_limit is None: set_concurrency_limit()
    return _concurrency_limit


def is_coro_func(obj, func_name: str = None) -> bool:
    """
    This is probably in the library elsewhere but returns bool
    based on if the function is a coro
    """
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))

    except Exception:
        return False



RT = TypeVar("RT")


async def amap_iterable(
    mapped_iterable: Union[Callable[[], Awaitable[Any]], Awaitable[Any], Coroutine[Any, Any, Any], Callable[[], Any]],
    concurrency_limit: Optional[int] = None,
    return_when: Optional[str] = 'FIRST_COMPLETED',
):
    """
    Limit the concurrency of an iterable

    Args:
        mapped_iterable (Union[Callable[[], Awaitable[Any]], Awaitable[Any], Coroutine[Any, Any, Any], Callable[[], Any]]): The iterable to limit the concurrency of
        limit (Optional[int], optional): The limit of the concurrency. Defaults to None.
        return_when (Optional[ReturnWhenType], optional): The return when type. Defaults to ReturnWhenType.FIRST_COMPLETED.
    
    Yields:
        [type]: [description]
    """
    try:
        iterable = aiter(mapped_iterable)
        is_async = True
    except (TypeError, AttributeError):
        iterable = iter(mapped_iterable)
        is_async = False
    
    iterable_ended: bool = False
    pending = set()
    concurrency_limit = get_concurrency_limit() if concurrency_limit is None else concurrency_limit
    return_when = getattr(asyncio, return_when) if isinstance(return_when, str) else return_when

    while pending or not iterable_ended:
        while len(pending) < concurrency_limit and not iterable_ended:
            try: iter_item = await anext(iterable) if is_async else next(iterable)
            except StopAsyncIteration if is_async else StopIteration:
                iterable_ended = True
            else: pending.add(asyncio.ensure_future(iter_item))

        if not pending: return
        done, pending = await asyncio.wait(pending,  return_when = return_when)
        while done: yield done.pop()



class ThreadPool(abc.ABC):
    """
    Pooler Base Class
    """

    allow_task_completion: Optional[bool] = True
    register_exit: Optional[bool] = True

    def __init__(
        self,
        max_workers: Optional[int] = None,
        allow_task_completion: Optional[bool] = None,
        register_exit: Optional[bool] = None,
        **kwargs
    ):
        """
        Initializes the Pooler
        """
        if allow_task_completion is not None: self.allow_task_completion = allow_task_completion
        if register_exit is not None: self.register_exit = register_exit
        if max_workers is None:
            max_workers = int(os.getenv("MAX_WORKERS", os.cpu_count()))
        
        self.max_workers = max_workers
        self.tasks: Optional[Set[asyncio.Task]] = set()
        self._pool: Optional[futures.ThreadPoolExecutor] = None
        self._ppool: Optional[futures.ProcessPoolExecutor] = None
        
        self._kwargs = kwargs

        if self.register_exit:
            import atexit
            atexit.register(self.on_exit)

    @staticmethod
    def is_coro(obj: Any) -> bool:
        """
        Checks if an object is a coroutine function
        """
        return is_coro_func(obj)
    
    def ensure_coro(self, func: Callable[..., RT]) -> Callable[..., Awaitable[RT]]:
        """
        Ensure that the function is a coroutine
        """
        if asyncio.iscoroutinefunction(func): return func
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            return await self.arun(func, *args, **kwargs)
        return inner
    
    def ensure_coro_function(self, func: Callable[..., RT]) -> Callable[..., Awaitable[RT]]:
        """
        Ensure that the function is a coroutine
        - Adds contextvars
        """
        if asyncio.iscoroutinefunction(func): return func
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            loop = asyncio.get_running_loop()
            ctx = contextvars.copy_context()
            return await loop.run_in_executor(
                executor = self.pool, 
                func = lambda: ctx.run(func, *args, **kwargs)
            )
        return inner
    
    @property
    def pool(self) -> futures.ThreadPoolExecutor:
        """
        Returns the ThreadPoolExecutor
        """
        if self._pool is None:
            self._pool = futures.ThreadPoolExecutor(max_workers = self.max_workers)
        return self._pool
    
    @property
    def ppool(self) -> futures.ProcessPoolExecutor:
        """
        Returns the ProcessPoolExecutor
        """
        if self._ppool is None:
            self._ppool = futures.ProcessPoolExecutor(max_workers = self.max_workers)
        return self._ppool
    
    @property
    def in_async_loop(self) -> bool:
        """
        Returns True if the current thread is in an async loop
        """
        try:
            return asyncio.get_running_loop() is not None
        except RuntimeError:
            return False
    
    def get_pool(
        self, 
        num_workers: Optional[int] = None, 
        process_pool: bool = False
    ) -> futures.Executor:
        """
        Returns the ThreadPoolExecutor or ProcessPoolExecutor
        """
        pool_cls = futures.ProcessPoolExecutor if process_pool else futures.ThreadPoolExecutor
        if num_workers is None: num_workers = self.max_workers
        return pool_cls(max_workers = num_workers)

    def add_task(
        self, 
        task: asyncio.Task, 
        callback: Optional[Callable] = None,
        callback_args: Optional[Tuple] = None,
        callback_kwargs: Optional[Dict] = None
    ):
        """
        Adds a task to the current tasks
        """
        self.tasks.add(task)
        if callback is not None:
            if callback_args or callback_kwargs:
                callback_args = callback_args or ()
                callback_kwargs = callback_kwargs or {}
                callback = functools.partial(callback, *callback_args, **callback_kwargs)
            task.add_done_callback(callback)
        task.add_done_callback(self.tasks.discard)


    def on_exit(self):
        """
        Cleans up the ThreadPoolExecutor and Tasks
        """
        for task in self.tasks:
            task.cancel()
        if self._pool is not None: self._pool.shutdown(wait = self.allow_task_completion, cancel_futures = not self.allow_task_completion)
        if self._ppool is not None: self._ppool.shutdown(wait = self.allow_task_completion, cancel_futures = not self.allow_task_completion)

    """
    Core
    """

    def run(self, func: Coroutine[RT], *args, **kwargs) -> RT:
        """
        Runs an Async Function as a Sync Function
        """
        current_async_module = getattr(threadlocals, "current_async_module", None)
        partial_f = functools.partial(func, *args, **kwargs)
        if current_async_module is None:
            return anyio.run(partial_f)
        return anyio.from_thread.run(partial_f)
    

    async def arun(self, func: Callable[..., RT], *args, **kwargs) -> RT:
        """
        Runs a Sync Function as an Async Function
        """
        blocking = functools.partial(func, *args, **kwargs)
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(self.pool, blocking)
    
    async def asyncish(self, func: Callable[..., RT], *args, **kwargs) -> RT:
        """
        Runs a Function as an Async Function if it is an Async Function
        otherwise wraps it around `arun`
        """
        if is_coro_func(func): return await func(*args, **kwargs)
        return await self.arun(func, *args, **kwargs)
    
    run_async = arun
    run_sync = run

    """
    Background Tasks
    """    

    def threadpool_task(
        self, 
        func: Callable, 
        *args, 
        task_callback: Optional[Callable[..., RT]] = None, 
        task_callback_args: Optional[Tuple] = None,
        task_callback_kwargs: Optional[Dict] = None,
        **kwargs
    ) -> futures.Future[RT]:
        """
        Creates a threadpool task
        """
        task = self.pool.submit(func, *args, **kwargs)
        self.add_task(task, task_callback, callback_args=task_callback_args, callback_kwargs=task_callback_kwargs)
        return task

    def background_task(
        self, 
        func: Callable[..., RT],
        *args, 
        task_callback: Optional[Callable] = None, 
        task_callback_args: Optional[Tuple] = None,
        task_callback_kwargs: Optional[Dict] = None,
        **kwargs
    ) -> Awaitable[RT]:
        """
        Creates a background task
        """
        if inspect.isawaitable(func): task = asyncio.create_task(func)
        else: task = asyncio.create_task(self.asyncish(func, *args, **kwargs))
        self.add_task(task, task_callback, callback_args = task_callback_args, callback_kwargs=task_callback_kwargs)
        return task
    
    def background(
        self, 
        func: Callable[..., RT], 
        *args, 
        task_callback: Optional[Callable] = None, 
        task_callback_args: Optional[Tuple] = None,
        task_callback_kwargs: Optional[Dict] = None,
        **kwargs
    ) -> Awaitable[RT]:
        """
        Runs a function in the background.
        If the current thread is in an async loop, it runs the function as an async function.
        Otherwise, it runs the function as a sync function in the threadpool.

        Returns a `asyncio.Task` if the current thread is in an async loop.
        Otherwise, it returns a `futures.Future`.
        """
        method = self.background_task if self.in_async_loop else self.threadpool_task
        return method(func, *args, task_callback = task_callback, task_callback_args = task_callback_args, task_callback_kwargs = task_callback_kwargs, **kwargs)

    create_threadpool_task = threadpool_task
    create_background_task = background_task
    create_background = background

    """
    Iterators
    """

    def map(
        self,
        func: Callable[..., RT],
        iterable: Iterable[Any],
        *args,
        return_ordered: Optional[bool] = True,
        use_process_pool: Optional[bool] = False, 
        **kwargs
    ) -> List[RT]:  # sourcery skip: assign-if-exp
        """
        Iterates over an iterable and runs a function on each item in the iterable.
        """
        num_workers = kwargs.pop('num_workers', None)
        partial_func = functools.partial(func, *args, **kwargs)
        with self.get_pool(num_workers = num_workers, process_pool = use_process_pool) as executor:
            _futures = [
                executor.submit(partial_func, item) for item in iterable
            ]
            if not return_ordered: return [f.result() for f in futures.as_completed(_futures)]
            futures.wait(_futures)
            return [f.result() for f in _futures]
    

    def iterate(
        self,
        func: Callable[..., RT],
        iterable: Iterable[Any],
        *args,
        use_process_pool: Optional[bool] = False, 
        return_ordered: Optional[bool] = True,
        **kwargs
    ) -> Generator[RT, None, None]:  # sourcery skip: assign-if-exp
        """
        Iterates over an iterable and runs a function on each item in the iterable.
        """
        num_workers = kwargs.pop('num_workers', None)
        partial_func = functools.partial(func, *args, **kwargs)
        with self.get_pool(num_workers = num_workers, process_pool = use_process_pool) as executor:
            _futures = [
                executor.submit(partial_func, item) for item in iterable
            ]
            if not return_ordered:
                for f in futures.as_completed(_futures):
                    yield f.result()
            else:
                futures.wait(_futures)
                for f in _futures:
                    yield f.result()
    
    async def amap(
        self,
        func: Callable[..., Awaitable[RT]],
        iterable: Iterable[Any], 
        *args,
        return_ordered: Optional[bool] = True,
        concurrency_limit: Optional[int] = None,
        **kwargs,
    ) -> List[RT]:
        """
        Creates an Async Generator that iterates over an iterable and runs a function on each item in the iterable.
        """
        return_when = kwargs.pop('return_when', 'ALL_COMPLETED' if return_ordered else 'FIRST_COMPLETED')
        concurrency_limit = kwargs.pop('limit', concurrency_limit)
        func = self.ensure_coro(func)
        partial = functools.partial(func, *args, **kwargs)
        try: mapped_iterable = map(partial, iterable)
        except TypeError: mapped_iterable = (partial(x) async for x in iterable)
        results = []
        async for result in amap_iterable(
            mapped_iterable, 
            return_when = return_when, 
            concurrency_limit = concurrency_limit
        ):
            results.append(await result)
        return results
    
    async def aiterate(
        self,
        func: Callable[..., Awaitable[RT]],
        iterable: Iterable[Any], 
        *args,
        return_ordered: Optional[bool] = True,
        concurrency_limit: Optional[int] = None,
        **kwargs,
    ) -> AsyncGenerator[RT, None]:
        """
        Creates an Async Generator that iterates over an iterable and runs a function on each item in the iterable.
        """
        return_when = kwargs.pop('return_when', 'ALL_COMPLETED' if return_ordered else 'FIRST_COMPLETED')
        concurrency_limit = kwargs.pop('limit', concurrency_limit)
        func = self.ensure_coro(func)
        partial = functools.partial(func, *args, **kwargs)
        try: mapped_iterable = map(partial, iterable)
        except TypeError: mapped_iterable = (partial(x) async for x in iterable)
        async for result in amap_iterable(
            mapped_iterable, 
            return_when = return_when, 
            concurrency_limit = concurrency_limit
        ):
            yield await result
    
    sync_map = map
    sync_iterate = iterate
    async_map = amap
    async_iterate = aiterate

    """
    CMD
    """

    @staticmethod
    def cmd(
        command: Union[List[str], str], 
        shell: bool = True, 
        raise_error: bool = True, 
        **kwargs
    ):
        if isinstance(command, list): command = " ".join(command)
        try:
            out = subprocess.check_output(command, shell=shell, **kwargs)
            if isinstance(out, bytes): out = out.decode('utf8')
            return out.strip()
        except Exception as e:
            if not raise_error: return ""
            raise e
    
    @staticmethod
    async def acmd(
        command: Union[str, List[str]], 
        output_only: bool = True, 
        stdout = asyncio.subprocess.PIPE, 
        stderr = asyncio.subprocess.PIPE, 
        output_encoding: str = 'UTF-8', 
        output_errors: str = 'ignore', 
        *args,
        **kwargs
    ) -> Union[str, asyncio.subprocess.Process]:
        """
        Executes a Shell command using `asyncio.subprocess.create_subprocess_shell`

        Returns str if output_only else `asyncio.subprocess.Process`
        """
        if isinstance(command, list): command = ' '.join(command)
        p = await asyncio.subprocess.create_subprocess_shell(command, *args, stdout = stdout, stderr = stderr, **kwargs)
        if not output_only: return p
        stdout, _ = await p.communicate()
        return stdout.decode(encoding = output_encoding, errors = output_errors).strip()
    
    run_command = cmd
    async_run_command = acmd
    


ThreadPooler: ThreadPool = ProxyObject(obj_cls = ThreadPool)
ThreadPoolV2 = ThreadPooler


def ensure_coro(
    func: Callable[..., Any]
) -> Callable[..., Awaitable[Any]]:
    """
    Ensure that the function is a coroutine
    """
    if asyncio.iscoroutinefunction(func): return func
    @functools.wraps(func)
    async def inner(*args, **kwargs):
        return await ThreadPooler.arun(func, *args, **kwargs)
    return inner




from __future__ import annotations

from typing import Sequence

from lazyops.libs.dbinit.base import Engine, TextClause, create_engine, text
from lazyops.libs.dbinit.data_structures.grant_to import GrantTo
from lazyops.libs.dbinit.data_structures.privileges import Privilege
from lazyops.libs.dbinit.entities.cluster_entity import ClusterEntity
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.entities.role import Role
from lazyops.libs.dbinit.mixins.grantable import Grantable


class Database(ClusterEntity, Grantable):
    """
    Represents a Postgres `Database <https://www.postgresql.org/docs/current/managing-databases.html>`_.
    """

    def __init__(
        self,
        name: str,
        depends_on: Sequence[Entity] | None = None,
        check_if_exists: bool | None = None,
        owner: Role | None = None,
        template: str | None = None,
        # encoding: str | None = None,
        # strategy: str | None = None,
        # locale: str | None = None,
        # lc_collate: str | None = None,
        # lc_ctype: str | None = None,
        # icu_locale: str | None = None,
        # locale_provider: str | None = None,
        # collation_version: str | None = None,
        # tablespace: str | None = None,
        allow_connections: bool | None = None,
        connection_limit: int | None = None,
        is_template: bool | None = None,
        # oid: str | None = None,
        grants: Sequence[GrantTo] | None = None,
    ):
        """
        All __init__ params correspond to CREATE DATABASE arguments and options, see
        `official Postgres documentation <https://www.postgresql.org/docs/current/sql-createdatabase.html>`_.


        :param name: Unique name of the Database. Must be unique across the cluster.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        :param owner: The :class:`lazyops.libs.dbinit.entitites.Role` who will own this database. Postgres defaults to the user executing the command.
        :param template: The name of the template from which to create the new database. Postgres defaults to template1.
        :param allow_connections: Flag to allow connections to this database. Postgres defaults to `True`.
        :param connection_limit: Number of concurrent connections that can be made to this database. Postgres defaults to -1, which means no limit.
        :param is_template: Flag to allow this database to be cloned by any user with CREATEDB privileges; if `False` (the default), then only superusers or the owner of the database can clone it.
        :param grants: Sequence of :class:`lazyops.libs.dbinit.data_structures.GrantTo` to specify privileges this database has in relation to specified roles.
        """
        self.owner = owner
        self.template = template
        # self.encoding = encoding
        # self.strategy = strategy
        # self.locale = locale
        # self.lc_collate = lc_collate
        # self.lc_ctype = lc_ctype
        # self.icu_locale = icu_locale
        # self.locale_provider = locale_provider
        # self.collation_version = collation_version
        # self.tablespace = tablespace
        self.allow_connections = allow_connections
        self.connection_limit = connection_limit
        self.is_template = is_template
        # self.oid = oid

        self._db_engine: Engine | None = None

        Grantable.__init__(self, name=name, grants=grants)
        ClusterEntity.__init__(self, name=name, depends_on=depends_on, check_if_exists=check_if_exists)

    def _create_statements(self) -> Sequence[TextClause]:
        # statement = f"CREATE DATABASE {self.name}"
        statement = f"CREATE DATABASE {self.encoded_name}"

        props = self._get_passed_args()

        # append the arguments to the sql statement
        for k, v in props.items():
            # case owner and type Role, use role.name
            if k == "owner" and isinstance(v, Role):
                # statement = f"{statement} OWNER={v.name}"
                statement = f"{statement} OWNER={v.encoded_name}"

            elif k != "grants":
                statement = f"{statement} {k.upper()}={v}"

        return [text(statement)]

    def _exists_statement(self) -> TextClause:
        # return text("SELECT EXISTS(SELECT 1 FROM pg_database WHERE datname=:db)").bindparams(db=self.encoded_name)
        return text("SELECT EXISTS(SELECT 1 FROM pg_database WHERE datname=:db)").bindparams(db=self.name)

    def _drop_statements(self) -> Sequence[TextClause]:
        statements = []
        if self.is_template:
            # cannot drop a template, must set is_template to false before drop
            statements.append(text(f"ALTER DATABASE {self.encoded_name} is_template false"))
            # statements.append(text(f"ALTER DATABASE {self.name} is_template false"))
        # statements.append(text(f"DROP DATABASE {self.name} (FORCE)"))
        statements.append(text(f"DROP DATABASE {self.encoded_name} (FORCE)"))
        return statements

    def _grant(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.engine(), statements=self._grant_statements(grantee=grantee, privileges=privileges)
        )

    def _grants_exist(self, grantee: Role, privileges: set[Privilege]) -> bool:
        rows = self._fetch_sql(engine=self.engine(), statement=self._grants_exist_statement())
        # filter to grantee and extract privileges
        try:
            existing_privileges = list(
                filter(None, [self._extract_privileges(acl=r[0], grantee=grantee) for r in rows])
            )[0]
            return privileges.issubset(existing_privileges)
        except IndexError:
            return False

    def _grants_exist_statement(self) -> TextClause:
        """
        The SQL statement that checks to see what grants exist.
        :return: A single :class:`sqlalchemy.TextClause` containing the SQL to check what grants exist on this entity.
        """
        return text("SELECT unnest(datacl) AS acl FROM pg_catalog.pg_database WHERE datname=:db_name").bindparams(
            # db_name=self.encoded_name
            db_name=self.name
        )

    def _revoke(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.engine(), statements=self._revoke_statements(grantee=grantee, privileges=privileges)
        )

    @staticmethod
    def _allowed_privileges() -> set[Privilege]:
        return {Privilege.CREATE, Privilege.CONNECT, Privilege.TEMPORARY, Privilege.ALL_PRIVILEGES}

    def db_engine(self) -> Engine:
        """
        Getter for this database's engine. Will create it if it is not yet set.
        :return: A `sqlalchemy.Engine` to connect to this database.
        """
        # database entities will reference this as the engine to use
        if not self._db_engine:
            # grab everything but db name from the cluster engine
            host = self.__class__.engine().url.host
            port = self.__class__.engine().url.port
            user = self.__class__.engine().url.username
            pw = self.__class__.engine().url.password

            # then create a new engine
            self._db_engine = create_engine(f"postgresql+psycopg://{user}:{pw}@{host}:{port}/{self.name}")
        return self._db_engine

    def __repr__(self):
        return f"<Database (name={self.name}, owner={self.owner})>"



from lazyops.types.models import BaseModel
from lazyops.types.static import VALID_REQUEST_KWARGS
from lazyops.types.classprops import lazyproperty
from lazyops.imports._aiohttpx import aiohttpx
from lazyops.imports._pydantic import get_pyd_fields, pyd_parse_obj, get_pyd_dict
from typing import Optional, Dict, Any, Tuple, Type, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny

__all__ = [
    'BaseResource',
    'ResourceType',
    'ResponseResource',
    'ResponseResourceType',
]


class BaseResource(BaseModel):

    @property
    def exclude_none(self):
        return True

    @property
    def exclude_fields(self):
        return {'exclude_none', 'exclude_fields'}

    @classmethod
    def parse_resource(
        cls,
        **kwargs
    ) -> Tuple[Type['BaseResource'], Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = [field.name for field in get_pyd_fields(cls)]
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields and k in VALID_REQUEST_KWARGS}
        # resource_obj = cls.parse_obj(resource_kwargs)
        resource_obj = pyd_parse_obj(cls, resource_kwargs)
        return resource_obj, return_kwargs


    @staticmethod
    def create_resource(
        resource: Type['BaseResource'],
        **kwargs
    ) -> Tuple[Type['BaseResource'], Dict]:
        """
        Extracts the resource from the kwargs and returns the resource 
        and the remaining kwargs
        """
        resource_fields = [field.name for field in get_pyd_fields(resource)]
        resource_kwargs = {k: v for k, v in kwargs.items() if k in resource_fields}
        return_kwargs = {k: v for k, v in kwargs.items() if k not in resource_fields and k in VALID_REQUEST_KWARGS}
        resource_obj = pyd_parse_obj(resource, resource_kwargs)
        return resource_obj, return_kwargs
    

    def dict(
        self,
        *,
        include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: Optional[bool] = None,
    ):
        """
        Returns the dict representation of the response
        """
        if exclude is None: exclude = set()
        if self.exclude_fields:
            exclude = set(exclude) | self.exclude_fields
        return get_pyd_dict(
            include = include, 
            exclude = exclude, 
            by_alias = by_alias,
            skip_defaults = skip_defaults,
            exclude_unset = exclude_unset,
            exclude_defaults = exclude_defaults,
            exclude_none = exclude_none if exclude_none is not None else self.exclude_none
        )
        # return super().dict(
        #     include = include, 
        #     exclude = exclude, 
        #     by_alias = by_alias,
        #     skip_defaults = skip_defaults,
        #     exclude_unset = exclude_unset,
        #     exclude_defaults = exclude_defaults,
        #     exclude_none = exclude_none if exclude_none is not None else self.exclude_none
        # )

    def parse_data(
        self,
        **kwargs
    ):
        """
        Parses the prediction response
        """
        pass


ResourceType = Type[BaseResource]


class ResponseResource(BaseResource):
    _input_obj: Optional[ResourceType] = None
    _response: Optional['aiohttpx.Response'] = None
    _streamed_data: Optional[str] = None

    @property
    def exclude_fields(self):
        return {'_input_obj', '_response', '_streamed_data', 'exclude_none', 'exclude_fields'}

    @lazyproperty
    def headers(self):
        return self._response.headers
    
    @lazyproperty
    def status_code(self):
        return self._response.status_code

    @lazyproperty
    def has_stream(self):
        return "text/event-stream" in self.headers.get("content-type", "")
    
    @lazyproperty
    def json_data(self) -> Dict[str, Any]:
        return self._response.json()

    @classmethod
    def parse_from(
        cls,
        input_obj: ResourceType,
        response: 'aiohttpx.Response',
        **kwargs
    ) -> 'ResponseResourceType':
        """
        Parses the response and returns the response object
        """
        resp = cls(
            _input_obj = input_obj,
            _response = response,
            **response.json()
        )
        resp.parse_data(**kwargs)
        return resp

ResponseResourceType = Type[ResponseResource]


from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Sequence

from lazyops.libs.dbinit.base import TextClause, text

from lazyops.libs.dbinit.data_structures.grant_to import GrantTo
from lazyops.libs.dbinit.data_structures.privileges import Privilege
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.exceptions import EntityExistsError, InvalidPrivilegeError

if TYPE_CHECKING:
    from lazyops.libs.dbinit.entities.role import Role


class Grantable(ABC):
    """
    Mixin for entities that can have privileges granted or revoked.
    """

    def __init__(self, name: str, grants: Sequence[GrantTo] | None = None):
        """
        :param name: Unique name of the grantable.
        :param grants: Sequence of grant definitions in the form of :class:`lazyops.libs.dbinit.data_structures.GrantTo`.
        """
        self.name = name
        self._grant_name = name
        if grants:
            self.grant(grants=grants)
    
    @property
    def encoded_grant_name(self) -> str:
        """
        Returns the encoded name of the grantable.
        """
        return f'"{self._grant_name}"' if '-' in self._grant_name or '_' in self._grant_name else self._grant_name

    def grant(self, grants: Sequence[GrantTo]) -> None:
        """
        Parses the provided grants, checks their validity, and stores them in the appropriate :class:`lazyops.libs.dbinit.entities.Role`.
        :param grants: Sequence of grant definitions in the form of :class:`lazyops.libs.dbinit.data_structures.GrantTo`.
        """
        for grant in grants:
            for grantee in grant.to:
                invalid_privileges = self._invalid_privileges(privileges=set(grant.privileges))
                if invalid_privileges:
                    formatted_invalid_privileges = ", ".join(invalid_privileges)
                    formatted_valid_privileges = ", ".join(self._allowed_privileges())
                    raise InvalidPrivilegeError(
                        f"Cannot grant the following privileges for database entity of "
                        f"type {self.__class__.__name__}: {formatted_invalid_privileges}. "
                        f"Valid privileges for this entity include: {formatted_valid_privileges}."
                    )
                else:
                    grantee.grants[self] = grantee.grants[self].union(set(grant.privileges))

    @abstractmethod
    def _exists(self) -> bool:
        """
        Check if this entity currently exists in the cluster.
        """
        pass

    def _safe_grant(self, grantee: Role, privileges: set[Privilege]) -> None:
        """
        Run an existence check before attempting to grant privileges.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to grant privileges to.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to grant.
        """
        if not self._exists():
            raise EntityExistsError(
                f"There is no {self.__class__.__name__} with the "
                f"name {self.name}. The {self.__class__.__name__} "
                f"must exist to grant privileges."
            )
        else:
            self._grant(grantee=grantee, privileges=privileges)

    @abstractmethod
    def _grant(self, grantee: Role, privileges: set[Privilege]) -> None:
        """
        Grant privileges on this entity to the grantee.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to grant privileges to.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to grant.
        """
        pass

    def _safe_revoke(self, grantee: Role, privileges: set[Privilege]) -> None:
        """
        Run an existence check before attempting to revoke privileges.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to revoke privileges from.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to revoke.
        """
        if not self._exists():
            raise EntityExistsError(
                f"There is no {self.__class__.__name__} with the "
                f"name {self.name}. The {self.__class__.__name__} "
                f"must exist to revoke privileges."
            )
        else:
            self._revoke(grantee=grantee, privileges=privileges)

    @abstractmethod
    def _grants_exist(self, grantee: Role, privileges: set[Privilege]) -> bool:
        """
        Check if these privileges are granted to the grantee.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to check access for.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to check on.
        """
        pass

    @abstractmethod
    def _revoke(self, grantee: Role, privileges: set[Privilege]) -> None:
        """
        Revoke privileges on this entity from the grantee.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to revoke privileges from.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to revoke.
        """
        pass

    def _grant_statements(self, grantee: Role, privileges: set[Privilege]) -> Sequence[TextClause]:
        """
        Generates a grant statement to commit via SQL.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to grant privileges to.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to grant.
        :return: A Sequence of :class:`sqlalchemy.TextClause` that represent the desired grant statements.
        """
        return [
            text(
                f"GRANT {self._format_privileges(privileges)} ON {self.__class__.__name__} {self.encoded_grant_name} TO {grantee.encoded_name}"
                # f"GRANT {self._format_privileges(privileges)} ON {self.__class__.__name__} {self._grant_name} TO {grantee.name}"
            )
        ]

    def _revoke_statements(self, grantee: Role, privileges: set[Privilege]) -> Sequence[TextClause]:
        """
        Generates a revoke statement to commit via SQL.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to revoke privileges from.
        :param privileges: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to revoke.
        :return: A Sequence of :class:`sqlalchemy.TextClause` that represent the desired revoke statements.
        """
        return [
            text(
                # f"REVOKE {self._format_privileges(privileges)} ON {self.__class__.__name__} {self._grant_name} FROM {grantee.name}"
                f"REVOKE {self._format_privileges(privileges)} ON {self.__class__.__name__} {self.encoded_grant_name} FROM {grantee.encoded_name}"
            )
        ]

    @staticmethod
    @abstractmethod
    def _allowed_privileges() -> set[Privilege]:
        """
        Define the set of :class:`lazyops.libs.dbinit.data_structures.Privilege` that are allowed for this entity.
        :return: The set of :class:`lazyops.libs.dbinit.data_structures.Privilege` that are allowed for this entity.
        """
        pass

    def _check_privileges(self, declared_privileges: set[Privilege], existing_privileges: set[Privilege]) -> bool:
        """
        Check the in-code declared privileges against the in-cluster existing privileges.
        :param declared_privileges: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` declared in code for this entity.
        :param existing_privileges: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` declared in cluster for this entity.
        :return: True if the declared privileges are a subset of the existing privileges. Accounts for ALL_PRIVILEGES.
        """
        if Privilege.ALL_PRIVILEGES in declared_privileges:
            declared = self._allowed_privileges()
            declared.discard(Privilege.ALL_PRIVILEGES)
        else:
            declared = declared_privileges

        return declared.issubset(existing_privileges)

    def _invalid_privileges(self, privileges: set[Privilege]) -> set[Privilege]:
        """
        Find all invalid privileges for this entity type.
        :param privileges: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to check for invalid entries.
        :return: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` that are invalid. Empty if all valid.
        """
        return privileges.difference(self._allowed_privileges())

    @staticmethod
    def _format_privileges(privileges: set[Privilege]) -> str:
        """
        Helper method that formats privileges for a SQL statement.
        :param privileges: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` to format.
        :return: A comma-separated list of the provided privileges as a string.
        """
        return ", ".join(privileges)

    @staticmethod
    def _fix_entity_order(grants: Sequence[GrantTo], target_entity: Entity) -> None:
        """
        If a grant relationship exists between a Grantable and a Role, the Role must exist prior to the Grantable.
        This method can be used by subclasses to ensure the entity order is correct, likely in the grant method.
        :param grants: A Sequence of :class:`lazyops.libs.dbinit.data_structures.GrantTo`.
        :param target_entity: The target to check against. Likely `self` or similar.
        """
        target_index = target_entity.entities.index(target_entity)
        for grant in grants:
            for grantee in grant.to:
                grantee_index = target_entity.entities.index(grantee)
                if grantee_index > target_index:
                    target_entity.entities.insert(target_index, target_entity.entities.pop(grantee_index))

    def _extract_privileges(self, acl: str, grantee: Role) -> set[Privilege]:
        """
        Extracts a set of :class:`lazyops.libs.dbinit.data_structures.Privilege` from a Postgres ACL statement.
        The expected format is: grantee=xxxx/grantor, where grantee and grantor are roles and the "x"s indicate
        potential privilege codes. If the grantee is absent, the grantee is PUBLIC.
        :param acl: Raw acl string from Postgres in the form of grantee=xxxx/grantor.
        :param grantee: The :class:`lazyops.libs.dbinit.entities.Role` to filter to.
        :return: A set of :class:`lazyops.libs.dbinit.data_structures.Privilege` that exist in cluster, granted to the grantee.
        """
        m = re.match(r"(\w*)=(\w*)\/(\w*)", acl)
        if m and m[1] == grantee.name:
            raw_privileges = m[2]
            return {self._code_to_privilege(code) for code in raw_privileges}
        return set()

    @staticmethod
    def _code_to_privilege(code: str) -> Privilege:
        """
        Wrapper around a dictionary to map from a letter code to a typed privilege.
        :param code: A letter code representing a privilege. See `Postgres docs <https://www.postgresql.org/docs/current/ddl-priv.html#PRIVILEGE-ABBREVS-TABLE>`_ for more.
        :return: A :class:`lazyops.libs.dbinit.data_structures.Privilege` corresponding to the provided letter code.
        """
        return {
            "r": Privilege.SELECT,
            "w": Privilege.UPDATE,
            "a": Privilege.INSERT,
            "d": Privilege.DELETE,
            "D": Privilege.TRUNCATE,
            "x": Privilege.REFERENCES,
            "t": Privilege.TRIGGER,
            "X": Privilege.EXECUTE,
            "U": Privilege.USAGE,
            "C": Privilege.CREATE,
            "c": Privilege.CONNECT,
            "T": Privilege.TEMPORARY,
        }[code]


class GrantableEntity(Grantable, Entity):
    """
    Convenience class that inherits from Grantable and Entity.
    """

    def grant(self, grants: Sequence[GrantTo]) -> None:
        super().grant(grants=grants)
        self._fix_entity_order(grants=grants, target_entity=self)



from .stoplight import OpenAPIStoplight
from .spec import (
    get_server_domain,
    patch_openapi_schema,
    create_openapi_schema_patch, 
    create_openapi_schema_by_role_function,
    OpenAPIRoleSpec,
    UserRole
)

"""
module_name = __package__.split('.')[0]

_openapi_spec_patches = {
    '/v1/services/create': {
        'source': 'Body_Create_v1_services_create',
        'schema': 'CreateService',
    },
}

_openapi_spec_exclude_schemas = [
    'BaseSchema',
    'BlockItem',
]

_included_paths = [
    '/v1/private/admin',
]


def description_patch(
    description: str,
    *args,
) -> str:

    description = description.split('page to authenticate.**\n\n', 1)[-1].strip()
    description = "## Private API Documentation\n" + description
    return description

role_specs = [
    OpenAPIRoleSpec(
        role = UserRole.ANON,
        excluded_tags = ['Private', 'Admin'],
        excluded_schemas = [
            'Private*',
        ],
    ),
    OpenAPIRoleSpec(
        role = UserRole.STAFF,
        included_paths = _included_paths,
        description_callable = description_patch,
    ),
    OpenAPIRoleSpec(
        role = UserRole.USER,
        excluded_tags = ['Admin'],
    ),
    OpenAPIRoleSpec(
        role = UserRole.SYSTEM,
        excluded_tags = ['Admin'],
    ),
    OpenAPIRoleSpec(
        role = UserRole.ADMIN,
        included_paths = _included_paths,
        description_callable = description_patch,
    ),
] 
patch_openapi_schema = create_openapi_schema_patch(
    module_name = module_name,
    schemas_patches = _openapi_spec_patches,
    excluded_schemas = _openapi_spec_exclude_schemas,
)
get_openapi_schema_by_role = create_openapi_schema_by_role_function(
    module_name = module_name,
    roles = role_specs,
    default_exclude_paths = _included_paths,
)


"""


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Sequence

from lazyops.libs.dbinit.base import  Engine, Row, TextClause


class SQLBase(ABC):
    @staticmethod
    def _commit_sql(engine: Engine, statements: Sequence[TextClause]) -> None:
        """
        Commits SQL statements to the database specified with the provided engine.
        :param engine: A :class:`sqlalchemy.Engine` for the target database.
        :param statements: A Sequence of :class:`sqlalchemy.TextClause` statements to commit.
        """
        with engine.connect() as conn:
            for statement in statements:
                conn.execution_options(isolation_level="AUTOCOMMIT").execute(statement)
                if hasattr(conn, 'commit'):
                    conn.commit()

    @staticmethod
    def _fetch_sql(engine: Engine, statement: TextClause) -> Sequence[Row[Any]]:
        """
        Fetches results from the database specified with the provided engine.
        :param engine: A :class:`sqlalchemy.Engine` for the target database.
        :param statement: A single :class:`sqlalchemy.TextClause` statement to fetch information from the database.
        :return: A Sequence of :class:`sqlalchemy.Row` that contain the results of the query provided.
        """
        with engine.connect() as conn:
            result = conn.execute(statement)
            return result.all()


class SQLCreatable(SQLBase):
    @abstractmethod
    def _create(self) -> None:
        """
        Create this entity in the cluster.
        """
        pass

    @abstractmethod
    def _exists(self) -> bool:
        """
        Check if this entity currently exists in the cluster.
        :return: True if it exists, False if it does not.
        """
        pass

    @abstractmethod
    def _drop(self) -> None:
        """
        Drop this entity from the cluster.
        """
        pass

    @abstractmethod
    def _create_statements(self) -> Sequence[TextClause]:
        """
        The SQL statements that create this entity.
        :return: A Sequence of :class:`sqlalchemy.TextClause` containing the SQL to create this entity.
        """
        pass

    @abstractmethod
    def _exists_statement(self) -> TextClause:
        """
        The SQL statement that checks to see if this entity exists.
        :return: A single :class:`sqlalchemy.TextClause` containing the SQL to check if this entity exists.
        """
        pass

    @abstractmethod
    def _drop_statements(self) -> Sequence[TextClause]:
        """
        The SQL statements that drop this entity.
        :return: A Sequence of :class:`sqlalchemy.TextClause` containing the SQL to drop this entity.
        """
        pass



from __future__ import annotations

from lazyops.types import BaseModel, Field
from typing import Optional, Dict, Any, Union, List

class SlackContext(BaseModel):
    """
    Holds the Slack Context
    """

    channels: Dict[str, str] = Field(default_factory = dict)
    users: Dict[str, str] = Field(default_factory = dict)
    user_info: Dict[str, Dict[str, Any]] = Field(default_factory = dict)
    user_lookup: Dict[str, str] = Field(default_factory = dict)
    uids: Dict[str, str] = Field(default_factory = dict)

    username_mapping: Optional[Dict[str, Union[str, List[str]]]] = Field(default_factory = dict)
    initialized: Optional[bool] = False
    
    def lookup_id(self, name: str) -> Optional[str]:
        """
        Lookup ID
        """
        return self.uids.get(name, self.users.get(name, self.channels.get(name)))




import time
from collections import OrderedDict, UserDict
from lazyops.utils.times import Timer
from lazyops.utils import logger

def test_native():
    logger.info('--- NATIVE ---')
    t = Timer()
    time.sleep(5)
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot

    time.sleep(5)
    logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    time.sleep(1)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')
    

def test_no_ckpt():

    logger.info('--- NO CKPT ---')

    t = Timer(auto_checkpoint=False)
    time.sleep(5)
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot

    time.sleep(5)
    logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    time.sleep(1)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')


def test_previous_duration():

    logger.info('-- Previous Duration ---')

    start_value = 240535.0
    t = Timer(start = start_value)
    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot
    logger.info('---')

    time.sleep(15)
    # logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    logger.info('---')
    time.sleep(5)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')

def test_previous_duration_short():
    start_value = 240535.0

    logger.info('-- Previous Duration (Short 0) ---')
    t = Timer(start = start_value, format_short= 0)
    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot
    logger.info('---')

    time.sleep(15)
    # logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    logger.info('---')
    time.sleep(5)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')

    logger.info('-- Previous Duration (Short 1) ---')
    t = Timer(start = start_value, format_short= 1)
    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot
    logger.info('---')

    time.sleep(15)
    # logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    logger.info('---')
    time.sleep(5)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')

    logger.info('-- Previous Duration (Short 2) ---')
    t = Timer(start = start_value, format_short= 2)
    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot
    logger.info('---')

    time.sleep(15)
    # logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    logger.info('---')
    time.sleep(5)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')

    logger.info('-- Previous Duration (Short 3) ---')

    t = Timer(start = start_value, format_short= 3)
    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')
    logger.info(f'Total: {t}')
    logger.info(f'Duration: {t.duration_s}', ) # Snapshot
    logger.info('---')

    time.sleep(15)
    # logger.info(f'Total: {t}')
    logger.info(f'Total: {t.total_s}') # Snapshot
    logger.info(f'Elapsed (Pre): {t.elapsed_s}')
    logger.info(f'Duration: {t.duration_s}')
    logger.info('---')
    time.sleep(5)
    logger.info(f'Elapsed (Post): {t.elapsed_s}')
    logger.info(f'Avg: {t.average_s}')
    logger.info(f'Total: {t}')


def test_op():

    import json
    start_value = 240535.0

    logger.info('-- Testing Operations ---')
    t = Timer(start = start_value)

    logger.info(f'Starting from: {t.pformat(start_value)}')
    logger.info(f'Start: {t.start}')

    t += 10.0
    logger.info(f'Add Total: {t}')

    t -= 5.0
    logger.info(f'Sub Total: {t}')

    t *= 2.0
    logger.info(f'Multiply Total: {t}')

    t /= 2.0
    logger.info(f'Divide Total: {t}')

    is_gt = t > start_value + 100.0
    logger.info(f'is_gt: {is_gt}')

    is_lt = t < start_value + 100.0
    logger.info(f'is_lt: {is_lt}')

    is_eq = t == start_value
    logger.info(f'is_eq: {is_eq}')
    # d = UserDict(t.dict())
    d = t.dict()
    # print(dir(d))
    print(t.__dict__)
    # from pydantic import ByteSize

    # b = ByteSize(1000)
    logger.info(f'Json Dump: {json.dumps(t)}')
    # logger.info(f'Json Dump: {json.dumps(b)}')



def run_tests():
    # test_native()
    # test_no_ckpt()
    # test_previous_duration()
    # test_previous_duration_short()
    test_op()

if __name__ == '__main__':
    run_tests()


from __future__ import annotations

"""
Timing Utils
"""

import abc
import time
from .logs import logger, null_logger
from typing import Optional, List, Dict, Any, Union


class Timer(abc.ABC, dict):
    start: Optional[float] = None

    def __init__(
        self,
        start: Optional[float] = None,
        
        auto_checkpoint: Optional[bool] = True,
        format_ms: Optional[bool] = False,
        format_pretty: Optional[bool] = True,
        format_short: Optional[int] = 0,
        verbose: bool = None,
    ):
        """
        Timer class for timing code

        Args:
            start (Optional[float], optional):
                The starting time of the timer. Defaults to None.
            auto_checkpoint (Optional[bool], optional):
                Whether to automatically add a checkpoint when accessing the duration.
                Defaults to True.
            format_ms (Optional[bool], optional):
                Whether to include milliseconds in the string representation of the duration.
                Defaults to False.
            format_pretty (Optional[bool], optional):
                Whether to use a pretty format for the string representation of the duration.
                Defaults to True.
            format_short (Optional[int], optional):
                0: Default format (full unit with spaces) 
                    2 days, 18 hours, 49 minutes, 15.01 seconds
                1: Pretty format (shortened unit with spaces)
                    2 days 18 hrs 49 mins 15.01 secs
                2: Short format (1 letter unit, single space)
                    2d 18h 49m 15s
                3: Short format (1 letter unit, no space)
                    2d18h49m15s
                Defaults to 1.

        """
        if not start:
            self.start = time.time()
        else:
            # If we're starting from a duration, we need to subtract the duration
            # from the current time
            curr_ts = time.time()
            if len(str(start)) == len(str(curr_ts)):
                self.start = float(start)
            else:
                self.start = curr_ts - float(start)
        
        self.durations: List[float] = []
        self.auto_checkpoint = auto_checkpoint
        self.verbose = verbose
        self.format_ms = format_ms
        self.format_pretty = format_pretty
        self.format_short = format_short
        # We do this to make the timer compatible with json.dumps
        # and initialize the dict
        # so that it won't be assumed to be empty
        self['_start'] = self.start

    def dformat(
        self,
        duration: float,
        pretty: bool = None,
        short: int = None,
        include_ms: bool = None,
        as_int: bool = False,
    ) -> Dict[str, Union[float, int]]:    # sourcery skip: low-code-quality
        """
        Formats a duration (secs) into a dict
        """
        pretty = pretty if pretty is not None else self.format_pretty
        short = short if short is not None else self.format_short
        include_ms = include_ms if include_ms is not None else self.format_ms
        if not pretty:
            unit = 'secs' if short else 'seconds'
            value = int(duration) if as_int else duration
            return {unit: value}
        data = {}
        if duration >= 86400:
            unit = 'd' if short > 1 else 'day'
            days = (duration // 86400)
            if short < 2 and days > 1: unit += 's'
            duration -= days * 86400
            data[unit] = int(days) if as_int else days
        if duration >= 3600:
            unit = 'hr' if short else 'hour'
            if short > 1: unit = unit[0]
            hours = (duration // 3600)
            if short < 2 and hours > 1: unit += 's'
            duration -= hours * 3600
            data[unit] = int(hours) if as_int else hours
        if duration >= 60:
            unit = 'min' if short else 'minute'
            if short > 1: unit = unit[0]
            minutes = (duration // 60)
            if short < 2 and minutes > 1: unit += 's'
            duration -= minutes * 60
            data[unit] = int(minutes) if as_int else minutes
        if duration >= 1:
            unit = 'sec' if short else 'second'
            if short > 1: unit = unit[0]
            if short < 2 and duration > 1: unit += 's'
            if include_ms:
                seconds = int(duration)
                duration -= seconds
                data[unit] = seconds
            elif short > 1:
                data[unit] = int(duration) if as_int else duration
            else:
                data[unit] = float(f'{duration:.2f}')
        if include_ms and duration > 0:
            unit = 'ms' if short else 'millisecond'
            milliseconds = int(duration * 1000)
            data[unit] = milliseconds
        return data

    def pformat(
        self, 
        duration: float,
        pretty: bool = None,
        short: int = None,
        include_ms: bool = None,
    ) -> str:  # sourcery skip: low-code-quality
        """
        Formats a duration (secs) into a string

        535003.0 -> 5 days, 5 hours, 50 minutes, 3 seconds
        3593.0 -> 59 minutes, 53 seconds
        """
        pretty = pretty if pretty is not None else self.format_pretty
        short = short if short is not None else self.format_short
        include_ms = include_ms if include_ms is not None else self.format_ms
        data = self.dformat(
            duration = duration,
            pretty = pretty,
            short = short,
            include_ms = include_ms,
            as_int = True
        )
        if not data: return '0 secs'
        sep = '' if short > 1 else ' '
        if short > 2: return ''.join([f'{v}{sep}{k}' for k, v in data.items()])
        return ' '.join([f'{v}{sep}{k}' for k, v in data.items()]) if short else ', '.join([f'{v}{sep}{k}' for k, v in data.items()])

    def checkpoint(self):
        """
        Adds a checkpoint to the timer
        """
        duration = time.time() - self.start
        self.durations.append(duration)
        self.start = time.time()

    def get_duration(self, checkpoint: Optional[bool] = None) -> float:
        """
        Returns the latest duration of the timer
        """
        checkpoint = checkpoint if checkpoint is not None else self.auto_checkpoint
        if checkpoint: self.checkpoint()
        return self.durations[-1] if self.durations else self.elapsed
    
    @property
    def duration(self) -> float:
        """
        Returns the latest duration of the timer
        """
        return self.get_duration()
        
    
    @property
    def duration_s(self) -> str:
        """
        Returns the latest duration of the timer as a string
        """
        # this is designed to prevent additional snapshotting
        # when duration is called and then duration_s is called
        # in logging.
        return self.pformat(self.get_duration(False))
    

    @property
    def total(self) -> float:
        """
        Returns the total duration of the timer
        """
        return sum(self.durations + [self.elapsed]) \
            if self.durations else self.elapsed
    
    @property
    def total_s(self) -> str:
        """
        Returns the total duration of the timer as a string
        """
        return self.pformat(self.total)
    
    def total_average(self, count: int) -> float:
        """
        Returns the average duration of the timer
        """
        return self.total / count
    
    def total_average_s(self, count: int) -> str:
        """
        Returns the average duration of the timer as a string
        """
        return self.pformat(self.total_average(count))
    
    @property
    def elapsed(self) -> float:
        """
        Returns the elapsed time since the timer was started
        Does not add a checkpoint
        """
        return time.time() - self.start
    
    @property
    def elapsed_s(self) -> str:
        """
        Returns the elapsed time since the timer was started as a string
        Does not add a checkpoint
        """
        return self.pformat(self.elapsed)
    
    def elapsed_average(self, count: int) -> float:
        """
        Returns the average duration of the timer
        """
        return self.elapsed / count
    
    def elapsed_average_s(self, count: int) -> str:
        """
        Returns the average duration of the timer as a string
        """
        return self.pformat(self.elapsed_average(count))

    @property
    def average(self) -> float:
        """
        Returns the average duration of the timer
        """
        if not self.durations: return self.elapsed
        return self.total / (len(self.durations)  + 1)
    
    @property
    def average_s(self) -> str:
        """
        Returns the average duration of the timer as a string
        """
        return self.pformat(self.average)

    @property
    def autologger(self):
        """
        Returns the logger to use
        """
        return logger if self.verbose else null_logger
    
    def __len__(self) -> int:
        """
        Returns the number of checkpoints
        """
        return len(self.durations)
    
    def __float__(self) -> float:
        """
        Returns the total duration of the timer
        """
        return self.total
    
    def __int__(self) -> int:
        """
        Returns the total duration of the timer
        """
        return int(self.total)
    
    def __str__(self) -> str:
        """
        Returns a string representation of the timer
        """
        return self.total_s
    
    def __repr__(self) -> str:
        """
        Returns a string representation of the timer
        """
        return f"Time({self.total_s})"
    
    def __getitem__(self, index: int) -> float:
        """
        Returns the duration at the given index
        """
        return self.durations[index] if self.durations else self.elapsed

    @property
    def data_dict(self) -> Dict[str, Any]:
        """
        Returns a dict representation of the timer
        """
        return {
            'duration': self.duration,
            'total': self.total,
            'elapsed': self.elapsed,
            'average': self.average,
            'data': self.dformat(self.total)
        }

    def dict(self, **kwargs) -> Dict[str, float]:
        """
        Returns a dict representation of the timer
        """
        return {
            'duration': self.duration,
            'total': self.total,
            'elapsed': self.elapsed,
            'average': self.average,
            'data': self.dformat(self.total)
        }
    
    def items(self) -> Dict[str, Union[str, float]]:
        """
        Returns a dict representation of the timer 
        that is compatible with json.dumps
        """
        return [
            ('duration', self.duration),
            ('total', self.total),
            ('elapsed', self.elapsed),
            ('average', self.average),
            ('data', self.dformat(self.total))
        ]

    
    def __eq__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is equal to the other value
        """
        if isinstance(other, Timer):
            return self.total == other.total
        return self.total == other
    
    def __ne__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is not equal to the other value
        """
        if isinstance(other, Timer):
            return self.total != other.total
        return self.total != other
    
    def __lt__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is less than the other value
        """
        if isinstance(other, Timer):
            return self.total < other.total
        return self.total < other
    
    def __le__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is less than or equal to the other value
        """
        if isinstance(other, Timer):
            return self.total <= other.total
        return self.total <= other
    
    def __gt__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is greater than the other value
        """
        if isinstance(other, Timer):
            return self.total > other.total
        return self.total > other
    
    def __ge__(self, other: Union[Timer, float, int]) -> bool:
        """
        Returns whether the timer is greater than or equal to the other value
        """
        if isinstance(other, Timer):
            return self.total >= other.total
        return self.total >= other
    
    
    def __add__(self, other: Union[Timer, float, int]) -> float:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            return self.total + other.total
        return self.total + other
    
    def __sub__(self, other: Union[Timer, float, int]) -> float:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            return self.total - other.total
        return self.total - other
    
    def __mul__(self, other: Union[Timer, float, int]) -> float:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            return self.total * other.total
        return self.total * other
    
    def __truediv__(self, other: Union[Timer, float, int]) -> float:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            return self.total / other.total
        return self.total / other
    
    def __floordiv__(self, other: Union[Timer, float, int]) -> float:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            return self.total // other.total
        return self.total // other
    

    """
    Mutations
    """
    
    def __iadd__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend(other.durations)
        else:
            self.durations.append(other)
        return self
    
    def __isub__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend([-d for d in other.durations])
        else:
            self.durations.append(-other)
        return self
    
    def __imul__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend([d * other.total for d in self.durations])
        else:
            self.durations.append(other * self.total)
        return self
    
    def __itruediv__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend([d / other.total for d in self.durations])
        else:
            self.durations.append(self.total / other)
        return self
    
    def __ifloordiv__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend([d // other.total for d in self.durations])
        else:
            self.durations.append(self.total // other)
        return self
    
    def __imod__(self, other: Union[Timer, float, int]) -> Timer:
        """
        Returns the total duration of the timer
        """
        if isinstance(other, Timer):
            self.durations.extend([d % other.total for d in self.durations])
        else:
            self.durations.append(self.total % other)
        return self
    



from __future__ import absolute_import

"""
Base Configs
"""
import pathlib
import contextlib
from typing import Optional, Union, Dict, TYPE_CHECKING
from lazyops.types.models import BaseSettings
from lazyops.types.classprops import lazyproperty

from lazyops.imports._fileio import (
    File, 
    FileLike, 
    _fileio_available, 
)

from lazyops.imports._aiokeydb import (
    aiokeydb, 
    resolve_aiokeydb,
    _aiokeydb_available, 
)

if TYPE_CHECKING:
    from lazyops.configs.cloud import BotoSettings, AwsSettings, GcpSettings, MinioSettings
    from lazyops.configs.k8s import K8sSettings
    from lazyops.configs.client import ClientSettings


class DefaultSettings(BaseSettings):

    ## General Variables
    api_host: Optional[str] = None
    api_port: Optional[int] = None
    
    app_home: Optional[str] = None
    app_name: Optional[str] = None
    app_host: Optional[str] = None
    app_port: Optional[int] = None
    
    ## Directory Variables
    app_dir: Optional[Union[str, pathlib.Path]] = None
    data_dir: Optional[Union[str, pathlib.Path]] = None
    model_dir: Optional[Union[str, pathlib.Path]] = None

    ## Context Variables
    in_docker: Optional[bool] = None

    ## Logging Variables
    log_level: Optional[str] = None
    log_format: Optional[str] = None
    
    ## Debug Variables
    debug_enabled: Optional[bool] = None
    gpg_key: Optional[str] = None

    ## Multithreading Variables
    max_workers: Optional[int] = None

    ## Huggingface Variables
    hf_home: Optional[Union[str, pathlib.Path]] = None
    hf_token: Optional[str] = None
    huggingface_token: Optional[str] = None

    huggingface_hub_cache: Optional[Union[str, pathlib.Path]] = None
    transformers_cache: Optional[Union[str, pathlib.Path]] = None
    sentence_transformers_home: Optional[Union[str, pathlib.Path]] = None

    """
    Host Variables
    """
    @lazyproperty
    def host_name(self) -> str:
        from lazyops.utils.system import get_host_name
        return get_host_name()

    @lazyproperty
    def host_ip(self):
        from lazyops.utils.system import get_host_ip
        return get_host_ip()
    
    @lazyproperty
    def cpu_count(self) -> int:
        from lazyops.utils.system import get_cpu_count
        return get_cpu_count()
    
    @lazyproperty
    def num_workers(self) -> int:
        return self.max_workers or (self.cpu_count * 2)
    

    """
    Directories
    """

    @lazyproperty
    def home(self) -> str:
        return pathlib.Path.home().as_posix()
    
    @lazyproperty
    def home_path(self) -> FileLike:
        return File(self.home) if _fileio_available else pathlib.Path(self.home)
    
    @lazyproperty
    def context_dir(self) -> str:
        if self.in_colab: return '/content'
        return '/' if self.is_remote else pathlib.Path.cwd().as_posix()
    
    @lazyproperty
    def context_path(self) -> FileLike:
        return File(self.context_dir) if _fileio_available else pathlib.Path(self.context_dir)

    @lazyproperty
    def app_path(self) -> FileLike:
        app_dir = self.app_dir or self.context_path.joinpath('app')
        return File(app_dir) if _fileio_available else pathlib.Path(app_dir)
    
    @lazyproperty
    def data_path(self) -> FileLike:
        data_dir = self.data_dir or self.context_path.joinpath('data')
        return File(data_dir) if _fileio_available else pathlib.Path(data_dir)
    
    @lazyproperty
    def model_path(self) -> FileLike:
        model_dir = self.model_dir or self.data_path.joinpath('models')
        return File(model_dir) if _fileio_available else pathlib.Path(model_dir)
    

    """
    Context / Env
    """

    ## Torch Variables
    @lazyproperty
    def torch_device(self):
        from lazyops.utils.system import get_torch_device
        return get_torch_device()
    
    @lazyproperty
    def torch_device_name(self):
        from lazyops.utils.system import get_torch_device_name
        return get_torch_device_name()
    
    @lazyproperty
    def in_colab(self) -> bool:
        with contextlib.suppress(Exception):
            from importlib import import_module
            import_module('google.colab')
            return True
        return False
    
    @lazyproperty
    def in_container(self) -> bool:
        return any(
            [self.in_docker,
            self.in_app,
            self.gpg_key is not None,]
        )

    @lazyproperty
    def in_k8s(self) -> bool:
        return self.k8s.in_cluster
    
    @lazyproperty
    def in_api(self) -> bool:
        return self.api_host is not None and self.api_port is not None
    
    @lazyproperty
    def in_app(self) -> bool:
        return (self.app_host is not None and self.app_port is not None) or (self.app_home and self.app_home == '/app')
    
    @lazyproperty
    def is_local(self) -> bool:
        return not any([self.in_k8s, self.in_k8s, self.in_api, self.in_docker])
    
    @lazyproperty
    def is_remote(self) -> bool:
        return self.in_k8s or self.in_api or self.in_docker or self.in_container
    
    """
    Huggingface Variables
    """
    
    @lazyproperty
    def hf_home_path(self) -> FileLike:
        hf_home = self.hf_home or self.model_path.joinpath('.hf')
        return File(hf_home) if _fileio_available else pathlib.Path(hf_home)

    @lazyproperty
    def hf_hub_cache_path(self) -> FileLike:
        if self.huggingface_hub_cache:
            return File(self.huggingface_hub_cache) if _fileio_available else pathlib.Path(self.huggingface_hub_cache)
        return self.hf_home_path
    
    @lazyproperty
    def huggingface_path(self) -> pathlib.Path:
        return self.home_path.joinpath('.huggingface')
    
    @lazyproperty
    def huggingface_token_path(self) -> pathlib.Path:
        return self.huggingface_path.joinpath('token')
    

    @lazyproperty
    def huggingface_token_exists(self) -> bool:
        return self.huggingface_token_path.exists()
    
    @lazyproperty
    def huggingface_auth_token(self) -> str:
        return self.hf_token or self.huggingface_token

    """
    Nested Cloud Settings
    """

    @lazyproperty
    def boto(self) -> 'BotoSettings':
        from lazyops.configs.cloud import BotoSettings
        return BotoSettings()

    @lazyproperty
    def aws(self) -> 'AwsSettings':
        from lazyops.configs.cloud import AwsSettings
        return AwsSettings()
    
    @lazyproperty
    def gcp(self) -> 'GcpSettings':
        from lazyops.configs.cloud import GcpSettings
        return GcpSettings()
    
    @lazyproperty
    def minio(self) -> 'MinioSettings':
        from lazyops.configs.cloud import MinioSettings
        return MinioSettings()
    
    """
    Nested Settings
    """

    @lazyproperty
    def k8s(self) -> 'K8sSettings':
        from lazyops.configs.k8s import K8sSettings
        return K8sSettings()
    
    @lazyproperty
    def keydb(self) -> 'aiokeydb.KeyDBSettings':
        if not _aiokeydb_available:
            resolve_aiokeydb()
        return aiokeydb.KeyDBClient.get_settings()
    
    @lazyproperty
    def client(self) -> 'ClientSettings':
        from lazyops.configs.client import ClientSettings
        return ClientSettings()
    
    """
    Client Props / Methods
    """

    def get_client_headers(self) -> Dict[str, str]:
        return self.client.get_headers()

    @property
    def default_client_name(self):
        """
        Returns the default API client name
        """
        if self.in_k8s: return 'cluster'
        return 'local' if self.client.api_dev_mode else 'external'

    @lazyproperty
    def client_endpoint(self) -> str:
        """
        Returns the client endpoint
        """
        return self.client.endpoints[self.default_client_name] or self.client.endpoints['local']

    


"""
Resolver for Pydantic v1/v2 imports with additional helpers
"""


import typing
from lazyops.utils.imports import resolve_missing


# Handle v1/v2 of pydantic
try:
    from pydantic import validator as _validator
    from pydantic import model_validator as base_root_validator
    from pydantic import ConfigDict

    PYD_VERSION = 2

    def root_validator(*args, **kwargs):
        """
        v1 Compatible root validator
        """
        def decorator(func):
            _pre_kw = kwargs.pop('pre', None)
            kwargs['mode'] = 'before' if _pre_kw is True else ('after' if _pre_kw is False else kwargs.get('mode', 'wrap'))
            return base_root_validator(*args, **kwargs)(func)
        return decorator

    def pre_root_validator(*args, **kwargs):
        def decorator(func):
            return base_root_validator(*args, mode='before', **kwargs)(func)
        return decorator
    
    def validator(*args, **kwargs):
        def decorator(func):
            return _validator(*args, **kwargs)(classmethod(func))
        return decorator

except ImportError:
    from pydantic import root_validator, validator

    ConfigDict = typing.Dict[str, typing.Any]

    PYD_VERSION = 1

    def pre_root_validator(*args, **kwargs):
        def decorator(func):
            return root_validator(*args, pre=True, **kwargs)(func)
        return decorator



try:
    from pydantic_settings import BaseSettings

except ImportError:
    if PYD_VERSION == 2:
        resolve_missing('pydantic-settings', required = True)
        from pydantic_settings import BaseSettings
    else:
        from pydantic import BaseSettings

import inspect
import pkg_resources
from pathlib import Path

class BaseAppSettings(BaseSettings):
    """
    BaseSettings with additional helpers
    """

    @property
    def module_path(self) -> Path:
        """
        Gets the module root path

        https://stackoverflow.com/questions/25389095/python-get-path-of-root-project-structure
        """
        return Path(pkg_resources.get_distribution(self.module_name).location)

    @property
    def module_config_path(self) -> Path:
        """
        Returns the config module path
        """
        return Path(inspect.getfile(self.__class__)).parent

    @property
    def module_name(self) -> str:
        """
        Returns the module name
        """
        return self.__class__.__module__.split(".")[0]
    
    @property
    def module_version(self) -> str:
        """
        Returns the module version
        """
        return pkg_resources.get_distribution(self.module_name).version
    
    @property
    def module_pkg_name(self) -> str:
        """
        Returns the module pkg name
        
        {pkg}/src   -> src
        {pkg}/{pkg} -> {pkg}
        """
        config_path = self.module_config_path.as_posix()
        module_path = self.module_path.as_posix()
        return config_path.replace(module_path, "").strip().split("/", 2)[1]

    @property
    def in_k8s(self) -> bool:
        """
        Returns whether the app is running in kubernetes
        """
        from lazyops.utils.system import is_in_kubernetes
        return is_in_kubernetes()
    
    @property
    def host_name(self) -> str:
        """
        Returns the hostname
        """
        from lazyops.utils.system import get_host_name
        return get_host_name()




from pydantic import BaseModel
from pydantic.fields import FieldInfo


def get_pyd_dict(model: typing.Union[BaseModel, BaseSettings], **kwargs) -> typing.Dict[str, typing.Any]:
    """
    Get a dict from a pydantic model
    """
    if kwargs: kwargs = {k:v for k,v in kwargs.items() if v is not None}
    return model.model_dump(**kwargs) if PYD_VERSION == 2 else model.dict(**kwargs)

def get_pyd_fields_dict(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.Dict[str, FieldInfo]:
    """
    Get a dict of fields from a pydantic model
    """
    return model.model_fields if PYD_VERSION == 2 else model.__fields__

def get_pyd_field_names(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.List[str]:
    """
    Get a list of field names from a pydantic model
    """
    return list(get_pyd_fields_dict(model).keys())

def get_pyd_fields(model: typing.Type[typing.Union[BaseModel, BaseSettings]]) -> typing.List[FieldInfo]:
    """
    Get a list of fields from a pydantic model
    """
    return list(get_pyd_fields_dict(model).values())

def pyd_parse_obj(model: typing.Type[typing.Union[BaseModel, BaseSettings]], obj: typing.Any, **kwargs) -> typing.Union[BaseModel, BaseSettings]:
    """
    Parse an object into a pydantic model
    """
    return model.model_validate(obj, **kwargs) if PYD_VERSION == 2 else model.parse_obj(obj)

def get_pyd_schema(model: typing.Type[typing.Union[BaseModel, BaseSettings]], **kwargs) -> typing.Dict[str, typing.Any]:
    """
    Get a pydantic schema
    """
    return model.schema(**kwargs) if PYD_VERSION == 2 else model.model_json_schema(**kwargs)


# https://github.com/pydantic/pydantic/issues/6763
# Patch until `pydantic-core` gets updated

# if PYD_VERSION == 2:
#     import pkg_resources


#     # If the module is < 2.10.1, then we need to patch it
#     pyd_core_version = pkg_resources.get_distribution('pydantic-core').version.split('.')
#     pyd_core_version_major = int(pyd_core_version[0])
#     pyd_core_version_minor = int(pyd_core_version[1])
#     pyd_core_version_patch = int(pyd_core_version[2])

#     if pyd_core_version_minor < 10 or pyd_core_version_patch < 2:

#         import weakref
#         import pydantic
#         # from pydantic.type_adapter import SchemaSerializer
#         from pydantic._internal._model_construction import SchemaSerializer
#         # from pydantic.schema import SchemaSerializer

#         class PickleableSchemaSerializer:
#             def __init__(self, schema, core_config):
#                 self._schema = schema
#                 self._core_config = core_config
#                 self._schema_serializer = SchemaSerializer(self._schema, self._core_config)

#             def __reduce__(self):
#                 return PickleableSchemaSerializer, (self._schema, self._core_config)

#             def __getattr__(self, attr: str):
#                 return getattr(self._schema_serializer, attr)

#         # class PickleableSchemaSerializer(SchemaSerializer):
#         #     def __init__(self, schema, core_config):
#         #         self._schema = schema
#         #         self._core_config = core_config
#         #         # No need for `super().__init__()` because `SchemaSerializer` initialization happens in `__new__`.

#         #     def __reduce__(self):
#         #         return PickleableSchemaSerializer, (self._schema, self._core_config)


#         class WeakRefWrapper:
#             def __init__(self, obj: typing.Any):
#                 self._wr = None if obj is None else weakref.ref(obj)

#             def __reduce__(self):
#                 return WeakRefWrapper, (self(),)

#             def __call__(self) -> typing.Any:
#                 return None if self._wr is None else self._wr()
        
#         pydantic._internal._model_construction.SchemaSerializer = PickleableSchemaSerializer
#         pydantic._internal._dataclasses.SchemaSerializer = PickleableSchemaSerializer
#         pydantic.type_adapter.SchemaSerializer = PickleableSchemaSerializer

#         # Override all usages of `_PydanticWeakRef` (obviously not needed if we upstream the above wrapper):
#         pydantic._internal._model_construction._PydanticWeakRef = WeakRefWrapper

    
